# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Shell capability interface."""
from __future__ import absolute_import
import abc
from gazoo_device.capabilities.interfaces import capability_base

_DEVICE_CAPABILITY_NAME = "shell_capability"
_SHELL_CMD = u"{cmd};echo Return Code: $?\n"
_SHELL_REGEX = r"(.*)Return Code: (\d+)\n"
_TIMEOUT = 10
_TRIES = 2


class ShellBase(capability_base.CapabilityBase):
    """Abstract base class defining the API for the shell() method."""

    def __init__(self, send_and_expect, device_name, shell_cmd=_SHELL_CMD,
                 shell_regex=_SHELL_REGEX, tries=_TRIES, timeout=_TIMEOUT):
        """Initalize the SSH shell capability.

        Args:
            send_and_expect (method): bound send_and_expect method of the device class instance.
            device_name (str): name of the device using this capability.
            shell_cmd (str): return code wrapper around the shell command to execute.
            shell_regex (str): shell regex to use. Must contain two capturing groups:
                               one for the output and one for the return code.
            tries (int): how many times to try sending the shell command.
            timeout (float): shell timeout in seconds.
        """
        super(ShellBase, self).__init__(device_name=device_name)

        self._send_and_expect = send_and_expect
        self._shell_regex = shell_regex
        self._shell_cmd = shell_cmd
        self._tries = tries
        self._timeout = timeout

    @classmethod
    def get_capability_name(cls):
        """Override the default to avoid a name conflict with the "shell" method."""
        return _DEVICE_CAPABILITY_NAME

    @abc.abstractmethod
    def shell(self, command, command_name="shell", timeout=10, port=0,
              include_return_code=False):
        """Sends command and returns response and optionally return code.

        Args:
            command (str): Command to send to the device.
            command_name (str): Identifier for command.
            timeout (float): Time in seconds to wait for device to respond.
            port (int): Which port to send on, 0 or 1.
            include_return_code (bool): flag indicating return code should be returned.

        Raises:
            GazooDeviceError: if communication fails.

        Note:
            Can try multiple times as connection can sometimes fail.
            See the init args for setting the number of retry attempts.

        Returns:
            str: If include_return_code is False return the device response to the command.
            tuple: If include_return_code is True return the device response and return code.
        """

    @abc.abstractmethod
    def has_command(self, binary_name):
        """Returns if binary_name is installed on the device.

        Args:
            binary_name (str): name of the executable.

        Returns:
            bool: True if the executable is found on the device, False otherwise.
        """



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Classes and variables to share across usb_info modules."""
import inspect

ANDROID_NAMES = ['Android', 'Pixel']
CAMBRIONIX_NAME = 'FT230X Basic UART'


CAMBRIONIX_PORT_MAP = {
    'PP8S': {"1.1": 1,
             "1.2": 2,
             "1.3": 3,
             "1.4": 4,
             "2.1": 8,
             "2.2": 7,
             "2.3": 6,
             "2.4": 5},
    'PP15S': {"1.1": 4,
              "1.2": 5,
              "1.3": 6,
              "1.4": 7,
              "2.1": 15,
              "2.2": 14,
              "2.3": 13,
              "2.4": 12,
              "3.1": 11,
              "3.2": 10,
              "3.3": 9,
              "3.4": 8,
              "4.2": 1,
              "4.3": 2,
              "4.4": 3},
    'U16S': {"2": 1,
             "5": 2,
             "4.1": 4,
             "4.2": 3,
             "4.3": 5,
             "4.4": 6,
             "4.5": 8,
             "4.6": 7,
             "4.7": 9,
             "7.1": 11,
             "7.2": 13,
             "7.3": 14,
             "7.4": 15,
             "7.5": 16,
             "7.6": 10,
             "7.7": 12}
}


class UsbInfo(object):
    """Class for storing all info about usb devices in a system agnostic way."""
    _address = None
    _child_addresses = []
    _ftdi_interface = 0
    _manufacturer = None
    _disk = None
    _product_id = None
    _product_name = None
    _serial_number = None
    _usb_hub_port = None
    _usb_hub_address = None
    _vendor_id = None

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            if not hasattr(self, key):
                raise ValueError(
                    "{} is not an arg. Choose from {}".format(
                        key, self.get_properties()))
            setattr(self, "_" + key, value)

    def __repr__(self):
        a_dict = {}
        for attr in self.get_properties():
            a_dict[attr] = getattr(self, attr)
        return str(a_dict)

    @classmethod
    def get_properties(cls):
        """Returns all the public properties."""
        return [name for name, value in inspect.getmembers(cls)
                if isinstance(value, property) and not name.startswith("_")]

    @property
    def address(self):
        """Address of device, persistent serial port path or adb serial."""
        return self._address

    @address.setter
    def address(self, a_string):
        self._address = a_string

    @property
    def child_addresses(self):
        """If cambrionix, list of addresses that are attached to cambrionix."""
        return self._child_addresses

    @child_addresses.setter
    def child_addresses(self, a_list):
        self._child_addresses = a_list

    @property
    def disk(self):
        """When volume is available, this is the disk link."""
        return self._disk

    @disk.setter
    def disk(self, a_string):
        self._disk = a_string

    @property
    def ftdi_interface(self):
        """Int 0 through 3 representing the ftdi interface."""
        return self._ftdi_interface

    @ftdi_interface.setter
    def ftdi_interface(self, a_number):
        self._ftdi_interface = a_number

    @property
    def manufacturer(self):
        """String that is the manufacturer's name."""
        return self._manufacturer

    @manufacturer.setter
    def manufacturer(self, a_string):
        self._manufacturer = a_string

    @property
    def product_id(self):
        """Product ID in 4 digit form. IE 1201."""
        return self._product_id

    @product_id.setter
    def product_id(self, a_string):
        self._product_id = a_string

    @property
    def product_name(self):
        """Product Name as a string. This is usually where programmed FTDI names appear."""
        return self._product_name

    @product_name.setter
    def product_name(self, a_string):
        self._product_name = a_string

    @property
    def serial_number(self):
        """Serial number of the connection. This could be an FTDI cable or a device."""
        return self._serial_number

    @serial_number.setter
    def serial_number(self, a_string):
        self._serial_number = a_string

    @property
    def usb_hub_port(self):
        """Number of the usb port the connection is on, if applicable."""
        return self._usb_hub_port

    @usb_hub_port.setter
    def usb_hub_port(self, a_number):
        self._usb_hub_port = a_number

    @property
    def usb_hub_address(self):
        """Number of the usb port the connection is on, if applicable."""
        return self._usb_hub_address

    @usb_hub_address.setter
    def usb_hub_address(self, a_string):
        self._usb_hub_address = a_string

    @property
    def vendor_id(self):
        """Product ID in 4 digit form. IE 1301."""
        return self._vendor_id

    @vendor_id.setter
    def vendor_id(self, a_string):
        self._vendor_id = a_string



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of the TransportProcess class which builds on the SwitchboardProcess base class.

The TransportProcess is responsible for reading and writing bytes to and from
the transport class provided and according to the following assumptions:

    * Transport commands from the main process are received using the
      send_command() method.

    * Raw transport data is (optionally) detokenized.

    * Raw/detokenized data is queued in the expect queue provided.

    * Expect queue can be enabled/disabled by the toggle_expect() method.

    * A host system timestamp is added to every line of data and queued in the
      log queue provided.

    * The host system timestamp will be in the following format:
      "<YYYY-MM-DD hh:mm:ss.ssssss> "

    * Custom log messages received as commands will only be added to the log
      queue between full device log messages.
"""
from __future__ import absolute_import
import multiprocessing
import six
import six.moves.queue as queue
import time
from gazoo_device.switchboard import switchboard_process
from gazoo_device.switchboard import data_framer
from gazoo_device.switchboard import log_process
from gazoo_device.switchboard import transport_properties as props
from six.moves import range


CMD_TRANSPORT_CLOSE = "TRANSPORT_CLOSE"
CMD_TRANSPORT_OPEN = "TRANSPORT_OPEN"
CMD_TRANSPORT_WRITE = "TRANSPORT_WRITE"
CMD_TRANSPORT_SET = "TRANSPORT_SET"
CMD_TRANSPORT_SERIAL_FLUSH = "TRANSPORT_SERIAL_FLUSH"
CMD_TRANSPORT_SERIAL_XON = "TRANSPORT_SERIAL_XON"
CMD_TRANSPORT_SERIAL_XOFF = "TRANSPORT_SERIAL_XOFF"
CMD_TRANSPORT_SERIAL_BREAK_BYTE = "TRANSPORT_SERIAL_BREAK_BYTE"
CMD_TRANSPORT_JLINK_RESET = "TRANSPORT_JLINK_RESET"
CMD_TRANSPORT_JLINK_FLASH = "TRANSPORT_JLINK_FLASH"
PARTIAL_LINE_TIMEOUT = 0.1  # time in seconds before publishing partial lines

_MAX_WRITE_BYTES = 32
_MAX_READ_BYTES = 11520  # 115200 / 10
_READ_TIMEOUT = 0.01  # ((115200 / 10) / 100ms) = ~115 bytes per 10ms read
_VALID_COMMANDS = [
    CMD_TRANSPORT_CLOSE,
    CMD_TRANSPORT_OPEN,
    CMD_TRANSPORT_WRITE,
    CMD_TRANSPORT_SET
]
_VALID_SERIAL_COMMANDS = [
    CMD_TRANSPORT_SERIAL_FLUSH,
    CMD_TRANSPORT_SERIAL_XON,
    CMD_TRANSPORT_SERIAL_XOFF,
    CMD_TRANSPORT_SERIAL_BREAK_BYTE,
]
_VALID_JLINK_COMMANDS = [
    CMD_TRANSPORT_JLINK_FLASH,
    CMD_TRANSPORT_JLINK_RESET
]
_ALL_VALID_COMMANDS = _VALID_COMMANDS + _VALID_SERIAL_COMMANDS + _VALID_JLINK_COMMANDS


def _enqueue_command_writes(write_queue, command, max_write_bytes=_MAX_WRITE_BYTES):
    """Splits command into max_write_bytes chunks and puts them into queue provided.

    Args:
        write_queue (Queue): to put split command into
        command (str): bytes to be written to device
        max_write_bytes (int): chunks to split command into.

    Note:
        Splitting commands into smaller write chunks avoids overloading the
        input buffer of the device.
    """

    command_split_indexes = list(range(0, len(command), max_write_bytes)) + [len(command)]
    for index in range(1, len(command_split_indexes)):
        start_index = command_split_indexes[index - 1]
        end_index = command_split_indexes[index]
        write_queue.put(command[start_index:end_index])


class TransportProcess(switchboard_process.SwitchboardProcess):
    """A process which manages (writes to and reads from) a transport instance."""

    def __init__(self,
                 device_name,
                 mp_manager,
                 exception_queue,
                 command_queue,
                 log_queue,
                 transport,
                 raw_data_queue=None,
                 raw_data_id=0,
                 framer=None,
                 partial_line_timeout=PARTIAL_LINE_TIMEOUT,
                 read_timeout=_READ_TIMEOUT,
                 max_read_bytes=_MAX_READ_BYTES,
                 max_write_bytes=_MAX_WRITE_BYTES):
        """Initialize TransportProcess with the arguments provided.

        Args:
            device_name (str): name of device using this transport
            mp_manager (multiprocessing.Manager): object to use for creating Events
            exception_queue (Queue): to use for reporting exception traceback
                                     message from subprocess
            command_queue (Queue): to receive commands into
            log_queue (Queue): to write each log line with host stamp added
            transport (Transport): to use to receive and send raw data
            raw_data_queue (Queue): to put raw (if applicable, detokenized)
                                    data into when enabled.
            raw_data_id (int): unique identifier for data published by this
                               transport process to the raw_data_queue.
            framer (DataFramer): to use to frame raw data into partial and
                                 complete lines.
            partial_line_timeout (float): time in seconds to wait before adding
                                          partial lines to raw_data_queue and
                                          log_queue.
            read_timeout (float): time to wait in seconds for transport reads.
            max_read_bytes (int): to attempt to read on each transport read
                                  call.
            max_write_bytes (int): to attempt to write on each transport write
                                   call.
        """

        process_name = "{}-Transport{}".format(device_name, raw_data_id)
        super(TransportProcess, self).__init__(device_name,
                                               process_name,
                                               mp_manager,
                                               exception_queue,
                                               command_queue,
                                               valid_commands=_ALL_VALID_COMMANDS)
        self._buffered_unicode = u""
        self._can_reopen = True
        self._framer = framer or data_framer.NewlineFramer()
        self._log_queue = log_queue
        self._max_read_bytes = max_read_bytes
        self._max_write_bytes = max_write_bytes
        self._partial_line_timeout = partial_line_timeout
        self._partial_log_time = time.time()
        self._pending_writes = None
        self._raw_data_enabled = multiprocessing.Event()
        self._raw_data_id = raw_data_id
        self._raw_data_queue = raw_data_queue
        self._read_timeout = read_timeout
        self._transport = transport
        self._transport_open = multiprocessing.Event()

    def get_raw_data(self, timeout=None):
        """Returns raw data message from optional raw data queue.

        Args:
            timeout (float): time to wait in seconds for incoming message to arrive.

        Raises:
            RuntimeError: if optional expect_queue was not provided when class
                          was initialized

        Returns:
            tuple: A tuple containing the raw_data_id and raw data received from
                   the transport device when enabled or None if no data was
                   available within timeout specified.
        """

        if self._raw_data_queue is None:
            raise RuntimeError("Device {} can't retrieve raw data. "
                               "No queue provided".
                               format(self.device_name))

        return switchboard_process.get_message(self._raw_data_queue, timeout=timeout)

    def is_open(self):
        """Returns True if transport is currently open.

        Returns:
            bool: A boolean indicating transport is currently open.
        """

        return self._transport_open.is_set()

    def raw_data_enabled(self):
        """Returns raw data streaming state.

        Returns:
            bool: A boolean indicating raw data streaming is enabled
        """

        return self._raw_data_enabled.is_set()

    def toggle_raw_data(self):
        """Toggles raw data queue to be enabled/disabled.

        Raises:
            RuntimeError: if optional raw_data_queue was not provided when class
                          was initialized
        """

        if self._raw_data_queue is None:
            raise RuntimeError("Device {} can't enable raw data output. "
                               "No queue provided".
                               format(self.device_name))

        if self._raw_data_enabled.is_set():
            self._raw_data_enabled.clear()
        else:
            self._raw_data_enabled.set()

    def _close_transport(self):
        if self._transport:
            self._transport.close()
            self._can_reopen = False
            if not self._transport.is_open():
                self._transport_open.clear()

    def _do_work(self):
        """Performs transport work and feeding log messages to log_queue.

        Returns:
            bool: always returns True
        """

        command_message = switchboard_process.get_message(self._command_queue, timeout=0)
        if command_message:
            self._process_command_message(command_message)
        if self._transport.is_open():
            self._transport_open.set()
            self._transport_write()
            self._transport_read()
        else:
            self._transport_open.clear()
            if self._transport.get_property(props.AUTO_REOPEN, False) and self._can_reopen:
                self._open_transport()
            else:
                time.sleep(self._read_timeout)
        return True

    def _is_line_published(self, line):
        elapsed_time = time.time() - self._partial_log_time
        if line and (line[-1] == "\n" or elapsed_time > self._partial_line_timeout):
            # NEP-3223 deal with partial line being finished by subsequent line
            if line[-1] == "\n" and self._buffered_unicode:
                self._publish_line(self._buffered_unicode + line)
                self._buffered_unicode = u""
            else:
                self._publish_line(line)
            return True
        return False

    def _open_transport(self):
        self._transport.open()
        self._can_reopen = True
        if self._transport.is_open():
            self._transport_open.set()

    def _post_run_hook(self):
        """Close transport at end of process."""
        self._close_transport()

    def _pre_run_hook(self):
        """Setup variables and open transport at start of process.

        Returns:
            bool: always returns True
        """

        self._pending_writes = queue.Queue()
        if self._transport.get_property(props.OPEN_ON_START, True):
            self._open_transport()
        self._partial_log_time = time.time()
        return True

    def _process_command_message(self, command_message):
        """Processes command messages which are received as a (command, data) tuple.

        Args:
            command_message (tuple): as received from command queue

        Raises:
            RuntimeError: if command in command_message is unknown.
            ValueError: if data doesn't contain a 2-tuple for CMD_TRANSPORT_SET.
            AttributeError: if the transport doesn't support the called command (eg. calling
                CMD_TRANSPORT_SERIAL_FLUSH on a non-serial transport.
        """

        command, data = command_message
        if CMD_TRANSPORT_CLOSE == command:
            self._close_transport()
        elif CMD_TRANSPORT_OPEN == command:
            self._open_transport()
        elif CMD_TRANSPORT_WRITE == command:
            _enqueue_command_writes(self._pending_writes,
                                    data,
                                    max_write_bytes=self._max_write_bytes)
        elif CMD_TRANSPORT_SET == command:
            property_key, property_value = data
            self._transport.set_property(property_key, property_value)
        elif CMD_TRANSPORT_SERIAL_FLUSH == command:
            self._transport.flush_buffers()
        elif CMD_TRANSPORT_SERIAL_XON == command:
            self._transport.send_xon()
        elif CMD_TRANSPORT_SERIAL_XOFF == command:
            self._transport.send_xoff()
        elif CMD_TRANSPORT_SERIAL_BREAK_BYTE == command:
            self._transport.send_break_byte()
        elif CMD_TRANSPORT_JLINK_FLASH == command:
            self._transport.flash(data)
        elif CMD_TRANSPORT_JLINK_RESET == command:
            self._transport.reset()
        else:
            raise RuntimeError("Device {} received an unknown command {}.".
                               format(self.device_name, command))

    def _publish_line(self, line):
        if self._raw_data_enabled.is_set():
            switchboard_process.put_message(self._raw_data_queue,
                                            (self._raw_data_id, line),
                                            timeout=0)
        log_process.log_message(self._log_queue, line, self._raw_data_id)

    def _transport_read(self):
        """Reads and processing incoming bytes from transport.
        """

        bytes_in = self._transport.read(size=self._max_read_bytes,
                                        timeout=self._read_timeout)
        if bytes_in:
            if isinstance(bytes_in, six.binary_type):
                unicode_in = bytes_in.decode("utf-8", "replace")
            else:
                unicode_in = bytes_in
            unicode_in = self._buffered_unicode + unicode_in
            buffered_len = len(self._buffered_unicode)
            self._buffered_unicode = u""
            # Can't use splitlines here because some devices
            # emit multiple line return characters without corresponding line
            # feed characters which prevents desired line splitting behavior.
            for line in self._framer.get_lines(unicode_in, begin=buffered_len):
                self._partial_log_time = time.time()
                if not self._is_line_published(line):
                    self._buffered_unicode += line
        elif self._buffered_unicode:
            if self._is_line_published(self._buffered_unicode):
                self._buffered_unicode = u""

    def _transport_write(self):
        """Writes previously split commands into transport."""
        if not self._pending_writes.empty():
            self._transport.write(self._pending_writes.get())



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility module for HTTP GET and POST methods."""

from __future__ import absolute_import

import json
import os
import socket
from gazoo_device import gdm_logger
import requests
import urllib3.exceptions
from requests import adapters
from requests.packages.urllib3 import poolmanager

from six.moves import range
from six.moves import urllib
MAX_URL_RETRY = 3

logger = gdm_logger.get_gdm_logger()


class SSLAdapter(adapters.HTTPAdapter):
    """An HTTPS Transport Adapter that uses an arbitrary SSL version."""

    def __init__(self, ssl_version=None, **kwargs):
        self.ssl_version = ssl_version

        super(SSLAdapter, self).__init__(**kwargs)

    def init_poolmanager(self, connections, maxsize, block=False):
        self.poolmanager = poolmanager.PoolManager(num_pools=connections,
                                                   maxsize=maxsize,
                                                   block=block,
                                                   ssl_version=self.ssl_version)


def download_file(url, local_path):
    """Downfile file to local path.

    Args:
        url (str): url to file.
        local_path (str): location to download file to.
    Raises:
      RuntimeError: if unable to ping domain or to download file in three tries.
    """
    for _ in range(MAX_URL_RETRY):
        try:
            _validate_url_access(url)
            (_, header) = urllib.request.urlretrieve(url, local_path)
        except Exception as err:
            raise RuntimeError("Unable to download {} "
                               "Error: {!r}".
                               format(url, err))
        if "last-modified" in header and os.path.exists(local_path):
            logger.info("Successfully downloaded file to {}".format(local_path))
            return
    raise RuntimeError("Unable to download {} after {} tries.".
                       format(url, MAX_URL_RETRY))


def read_raw_html_page(url):
    """returns the raw code on the html page.

    Args:
       url (str): url for page.

    Returns:
       str: raw code on the html page.
    Raises:
      RuntimeError: if unable to ping domain or to download file in three tries.
    """
    code = 0
    for _ in range(MAX_URL_RETRY):
        try:
            _validate_url_access(url)
            response = urllib.request.urlopen(url)
            code = response.getcode()
            raw_page = response.read().decode("utf-8")
            if code == 200:
                # get all relative links to files
                return raw_page
            else:
                err = raw_page
        except urllib.error.HTTPError as err:
            code = err.code  # retry
            raw_page = err
    raise RuntimeError("Unable to get files at URL {}. HTTP return code {}: {}".
                       format(url, code, raw_page))


def send_http_get(url,
                  auth=None,
                  params=None,
                  data=None,
                  headers=None,
                  ssl_version=None,
                  valid_return_codes=None,
                  timeout=10,
                  tries=1):
    """Issues a HTTP GET and returns the response if the request is successful.

    Args:
        url (str): HTTP URL to which HTTP GET request needs to be sent
        auth (auth.AuthBase): HTTP authentication object (i.e. HTTPDigestAuth)
        params (dict): Parameters to send as a query string.
        data (dict): Data that is needed for this HTTP GET Request
        headers (dict): Headers that is needed for this HTTP GET Request
        ssl_version (int): SSL version to be used for secure http (https) requests.
                           For example, ssl.PROTOCOL_TLSv1_2.
        valid_return_codes (list): List of valid HTTP return codes.
        timeout (int): request timeout in seconds
        tries (int): how many times to try sending the request.

    Raises:
        RuntimeError: if response.status_code returned by requests.get() is not in
                      valid_return_codes
        TypeError: if headers is not a dictionary
        RequestException, HTTPError: raised if request fails and are converted into a RuntimeError.

    Returns:
        requests.Response: Response returned by requests.get
    """
    valid_return_codes = valid_return_codes or [200]

    if data and not isinstance(data, dict):
        raise TypeError("Expecting a dict value data param but received: {}".
                        format(type(data)))
    if headers and not isinstance(headers, dict):
        raise TypeError("Expecting a dict value in headers param but received: {}".
                        format(type(headers)))
    try:
        session = requests.Session()
        if ssl_version:
            session.mount('https://', SSLAdapter(ssl_version))
        if data:
            data = json.dumps(data)

        for attempt in range(tries):
            try:
                response = session.get(url,
                                       auth=auth,
                                       params=params,
                                       data=data,
                                       headers=headers,
                                       timeout=timeout,
                                       verify=False)
                break
            except (requests.exceptions.RequestException, urllib3.exceptions.HTTPError) as err:
                if attempt < tries - 1:
                    logger.debug("HTTP GET to URL {} with data {} failed. Retrying. Error: {!r}"
                                 .format(url, data, err))
                else:
                    raise

    except Exception as err:
        raise RuntimeError("HTTP GET to URL {} with data {} failed with error: {!r}".format(
            url, data, err))

    if response.status_code not in valid_return_codes:
        raise RuntimeError("HTTP GET to URL {} returned: {}".format(url, response.reason))

    return response


def send_http_post(url,
                   auth=None,
                   data=None,
                   headers=None,
                   json_data=None,
                   ssl_version=None,
                   valid_return_codes=None,
                   timeout=10,
                   tries=1):
    """Issues a HTTP POST and returns the response if the request is successful.

    Args:
        url (str): HTTP URL to which HTTP POST request needs to be sent
        auth (auth.AuthBase): HTTP authentication object (i.e. HTTPDigestAuth)
        data (dict): Data that is needed for this HTTP GET Request
        headers (dict): Headers needed that is needed for this HTTP POST Request
        json_data (dict): JSON data that is needed for this HTTP POST Request
        ssl_version (int): SSL version to be used for secure http (https) requests.
                           For example, ssl.PROTOCOL_TLSv1_2
        valid_return_codes (list): List of valid HTTP return codes.
        timeout (int): request timeout in seconds
        tries (int): how many times to try sending the request.

    Raises:
        RuntimeError: if response.status_code returned by requests.post() is not in
        valid_return_codes
        TypeError: if headers / json_data is not a dictionary or None.
        RequestException, HTTPError: raised if request fails and are converted into a RuntimeError.

    Returns:
        requests.Response: Response returned by requests.post
    """
    headers = headers or {}
    json_data = json_data or {}
    valid_return_codes = valid_return_codes or [200]

    if not isinstance(headers, dict) or not isinstance(json_data, dict):
        raise TypeError("Expecting a dict value in headers and json_data params but received: "
                        "{} {} respectively".format(type(headers), type(json_data)))

    try:
        session = requests.Session()
        if ssl_version:
            session.mount('https://', SSLAdapter(ssl_version))

        for attempt in range(tries):
            try:
                response = session.post(url,
                                        auth=auth,
                                        data=data,
                                        json=json_data,
                                        headers=headers,
                                        timeout=timeout,
                                        verify=False)
                break
            except (requests.exceptions.RequestException, urllib3.exceptions.HTTPError) as err:
                if attempt < tries - 1:
                    logger.debug("HTTP POST to URL {} with headers {}, data {} and json data {} "
                                 "failed. Retrying. Error: {!r}"
                                 .format(url, headers, data, json_data, err))
                else:
                    raise

    except Exception as err:
        raise RuntimeError("HTTP POST to URL {} with headers {}, data {} and json data {} failed "
                           "with error: {!r}".format(url, headers, data, json_data, err))

    if response.status_code not in valid_return_codes:
        raise RuntimeError("HTTP POST to URL {} with headers {}, data {} and json data {} "
                           "returned: {}".format(url, headers, data, json_data, response.reason))

    return response


def is_valid_ip_address(address, is_ipv6=False):
    """Checks if valid IP address.

    Args:
        address (str): IP address
        is_ipv6 (bool): if IP address is IPv6

    Returns:
        bool: True if valid IP else False
    """
    address_family = socket.AF_INET6 if is_ipv6 else socket.AF_INET
    if is_ipv6 and '%' in address:  # Check for link local Ipv6 address
        # Example: fe80::2e0:4cff:fe01:4d1%en0
        address = address.split('%')[0]
    try:
        socket.inet_pton(address_family, address)
    except socket.error:  # not a valid address
        return False
    return True


def _validate_url_access(url):
    """Validates url access.

    Args:
        url (str): Url to ping.

    Raises:
        RuntimeError: if unable to ping domain of url.
    """
    parsed_uri = urllib.parse.urlparse(url)
    prefix = '{uri.scheme}://{uri.netloc}/'.format(uri=parsed_uri)
    response_code = urllib.request.urlopen(prefix).getcode()
    if response_code != 200:
        raise RuntimeError(
            "Unable to reach domain {}. Returned code {}".format(
                prefix, response_code))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Detect Device."""
from __future__ import absolute_import
import os
import weakref
from gazoo_device import config
from gazoo_device import detect_criteria
from gazoo_device import gdm_logger
from gazoo_device.switchboard import communication_types


logger = gdm_logger.get_gdm_logger()


class DeviceDetector(object):
    """Class for detecting devices.

    Class scans systems for all connections, filters out known connections,
    then attempts to assign connection to right device classes and create
    entry in configs for persistent info (such as ID and serial_number)
    """

    def __init__(self, manager, log_directory,
                 supported_auxiliary_device_classes):
        """Initializes the device detector.

        Args:
            manager (Manager): instance of manager.
            log_directory (str): location of logs.
            supported_auxiliary_device_classes (list): list of auxiliary device classes.
        """
        self.manager_weakref = weakref.ref(manager)
        self.log_directory = log_directory
        self.auxiliary_classes = supported_auxiliary_device_classes
        self.persistent_configs = {
            "devices": self.manager_weakref().persistent_dict.copy(),
            "other_devices": self.manager_weakref().other_persistent_dict.copy()
        }
        self.options_configs = {
            "device_options": self.manager_weakref().options_dict.copy(),
            "other_device_options": self.manager_weakref().other_options_dict.copy()
        }
        self.known_connections = self._create_known_connections()

    def detect_all_new_devices(self, static_ips=None):
        """Finds all possible new connections and detects devices.

        Args:
          static_ips (list): static ips not otherwise detectable.

        Returns:
          tuple: dict of persistent props, dict of optional props
        """
        logger.info("\n##### Step 1/3: Detecting potential new communication addresses. #####\n")
        all_connections_dict = communication_types.detect_connections(static_ips)
        return self.detect_new_devices(all_connections_dict)

    def detect_new_devices(self, connections_dict):
        """Detects the devices for given connections.

        Args:
          connections_dict (dict): paths to connections grouped by type.

        Returns:
          tuple: persistent_configs, options_configs for all devices (new and old)

        --> Filters out known connections
        --> Determine device type
        --> Detects persistent info
        --> Updates persistent and optional configs
        --> Tracks errors and unidentified connections and Writes a summary
        """

        connections_dict = self._filter_out_known_connections(connections_dict,
                                                              self.known_connections)

        possible_device_tuples, errs, no_id_cons = self._identify_connection_device_class(
            connections_dict)
        new_names = []
        logger.info("\n##### Step 3/3: Extract Persistent Info from Detected Devices. #####\n")
        for device_class, connection in possible_device_tuples:
            try:
                name, persistent_props, optional_props = self._detect_get_info(
                    device_class, connection)
                new_names.append(name)
                self._add_to_configs(device_class, name, persistent_props, optional_props)
            except Exception as err:
                msg = "Error extracting info from {} {!r}. Err: {!r}".format(
                    device_class.DEVICE_TYPE, connection, err)
                errs.append(msg)
                no_id_cons.append(connection)

        self._print_summary(new_names, errs, no_id_cons)
        return self.persistent_configs, self.options_configs

    def _add_to_configs(self, device_class, name, persistent_props, optional_props):
        """Adds newly detect info to appropriate location in config.

        Args:
          device_class (device_class): device class
          name (str): name of device
          persistent_props (dict): dict of device's persistent props
          optional_props (dict): dict of device's optional props
        """

        if device_class in self.auxiliary_classes:
            ind = 1
            more_props = {key: None for key in config.HUB_OPTION_ATTRIBUTES}
        else:
            ind = 0
            more_props = {key: None for key in config.DEVICE_OPTION_ATTRIBUTES}

        # Add generic props to options
        optional_props.update(more_props)

        self.persistent_configs[config.DEVICES_KEYS[ind]][name] = persistent_props
        self.options_configs[config.OPTIONS_KEYS[ind]][name] = optional_props

    def _create_known_connections(self):
        """Returns all known connections.

        Returns:
           list: known connection paths.
        """
        known_connections = []

        for _, entry in self.persistent_configs['devices'].items():
            identifier = entry.get("console_port_name")
            if not identifier:
                # Use the IP address instead of the ADB port for ADB over IP
                identifier = entry.get("adb_serial").replace(":5555", "")
            known_connections.append(identifier)
        for _, entry in self.persistent_configs["other_devices"].items():
            identifier = entry.get("console_port_name")
            if not identifier:
                identifier = entry.get("hub_port_name")
            known_connections.append(identifier)

        return known_connections

    def _detect_get_info(self, device_class, connection):
        """Returns name, persistent and optional info from device communication.

        Args:
            device_class (type): device class with get_detection_info method.
            connection (str): path to path communication

        Returns:
            tuple: name, dict of persistent props, dict of options props

        Note:
           Any errors raised will be caught in parent method.
        """
        device_type = device_class.DEVICE_TYPE
        detect_file = self._get_detect_log_file(connection)
        device_config = {
            "persistent": {
                "console_port_name": connection,
                "device_type": device_type,
                "name": "\t{}_detect".format(device_type)},
            "options": {},
            "make_device_ready": "on"
        }
        logger.info(
            "Getting info from communication port {} for {}".format(
                connection,
                device_type))
        device = device_class(manager=self.manager_weakref(),
                              device_config=device_config,
                              log_directory=self.log_directory,
                              log_file_name=detect_file)
        try:
            device.make_device_ready()
            persistent_props, options_props = device.get_detection_info()
        finally:
            device.close()

        name = self._generate_name(device_type,
                                   persistent_props["serial_number"],
                                   device_class)
        persistent_props['name'] = name
        return name, persistent_props, options_props

    def _filter_out_known_connections(self, con_dict, known_cons):
        """Filters out already known connections.

        Args:
          con_dict (dict): dictionary of detected connections by type
          known_cons (list): list of known connections.

        Returns:
          dict: dictionary of unknown connections by type.
        """
        new_con_dict = {}
        for key, con_list in con_dict.items():
            new_con_dict[key] = [
                con for con in con_list
                if con.replace(u":5555", u"") not in known_cons]
            if new_con_dict[key]:
                logger.info(
                    "Found {} possible {} connections:".format(
                        len(new_con_dict[key]), key.lower().replace("_", " ")))
                logger.info(u"\t" + u"\n\t".join(new_con_dict[key]))
        return new_con_dict

    def _generate_name(self, device_type, serial_number, device_class):
        """Generates name from last four digits of serial_number.

        Args:
            device_type (str): device type.
            serial_number (str): device serial number.
            device_class (class): identifying class

        Returns:
            str: name of device

        Raises:
            RuntimeError: if serial number is too short.

        Note:
          If there is a conflict with a different device it will generate a longer name.
        """
        if len(serial_number) < 4:
            raise RuntimeError("{} serial number {} is too short.".format(
                device_type, serial_number))
        name = "{}-{}".format(device_type, serial_number[-4:].lower())

        # Check if name already exists
        if device_class in self.auxiliary_classes:
            key = "other_devices"
        else:
            key = "devices"
        if name in self.persistent_configs[key]:
            # Check if its the same device or not
            if serial_number != self.persistent_configs[key][name]["serial_number"]:
                name = "{}-{}".format(device_type, serial_number[-8:].lower())
        return name

    def _get_detect_log_file(self, address):
        name = address.replace("/", "_")
        return "{}_detect.txt".format(name)

    def _identify_connection_device_class(self, connections_dict):
        """Detects all connections' appropriate device class.

        Args:
            connections_dict (dict): dictionary of connections by communication type

        Returns:
          tuple: tuple of lists. list of (class, connection), errs, connections not id'd
        """
        possible_device_tuples = []
        errs = []
        no_id_cons = []
        logger.info("\n##### Step 2/3 Identify Device Type of Connections. #####\n")

        for key in sorted(connections_dict.keys()):
            if not connections_dict[key]:  # no connections of that type
                continue
            key_name = key.lower().replace("_", " ")
            logger.info("Identifying {} devices..".format(key_name))
            for connection in connections_dict[key]:
                detect_log = os.path.join(self.log_directory,
                                          self._get_detect_log_file(connection))
                matching_classes = detect_criteria.determine_device_class(
                    connection, key, detect_log, self.manager_weakref().create_switchboard)
                if len(matching_classes) > 1:
                    device_types = [device_class.DEVICE_TYPE for device_class in matching_classes]
                    logger.info("Warning: Multiple device classes matched connection {}."
                                " If this is not an issue with your device controller package, "
                                "open an issue against GDM. Returning {}".format(device_types,
                                                                                 device_types[0]))
                if matching_classes:
                    logger.info(
                        "\t{} is a {}.".format(
                            connection,
                            matching_classes[0].DEVICE_TYPE))
                    possible_device_tuples.append((matching_classes[0], connection))
                else:
                    logger.info(
                        "\t{} responses did not match a known device type.".format(connection))
                    no_id_cons.append(connection)
            logger.info("\t{} device_type detection complete.".format(key_name))
        return possible_device_tuples, errs, no_id_cons

    def _print_summary(self, names, errs, no_id_cons):
        """Prints summary of detection events.

        Args:
          names (list): list of detected names.
          errs (list): list of errors.
          no_id_cons (list): list of undetected connections.
        """
        logger.info("\n##### Detection Summary #####\n")
        logger.info("\t{} new devices detected:{}", len(names),
                    "\n\t\t" + "\n\t\t".join(names))
        if errs:
            logger.info("\n\t{} errors/warnings:{}",
                        len(errs),
                        "\n\t\t" + "\n\t\t".join(errs))
        if no_id_cons:
            logger.info("\n\t{} connections found but not detected:{}",
                        len(no_id_cons),
                        "\n\t\t" + "\n\t\t".join(no_id_cons))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Build Flasher capability interface."""
import abc
from gazoo_device.capabilities.interfaces import capability_base


class FlashBuildBase(capability_base.CapabilityBase):
    """Abstract base class for the build_flasher capability."""

    @abc.abstractmethod
    def download_build_file(self, remote_build_folder, local_folder):
        """Retrieves the build file(s) from the remote location and puts them in the local folder.

        Args:
            remote_build_folder (str): build folder in remote location.
            local_folder (str): local path to store build files.
        """

    @abc.abstractmethod
    def get_defaults(self):
        """Returns a dictionary of default build arguments.

        Returns:
            dict: the default build arguments for the device.
        """

    @abc.abstractmethod
    def get_firmware_version(self, build_args=None):
        """Returns the firmware version based on the build arguments.

        Args:
            build_args (dict): dictionary of build arguments.
                               If None, the default build arguments are used.

        Returns:
            str: firmware version
        """

    @abc.abstractmethod
    def get_remote_build_folder(self, build_args=None):
        """Uses the build arguments to determine the remote build folder.

        Args:
            build_args (dict): dictionary of build arguments.
                               If None, the default build arguments are used.

        Returns:
            str: the remote build folder path.
        """

    @abc.abstractmethod
    def extract_build_info(self, build_args=None):
        """Converts the provided build arguments into information about the build.

        Args:
            build_args (dict): dictionary of build arguments.
                               If None, the default build arguments are used.

        Returns:
            dict: dictionary of build information including absolute_remote_path, build_type,
                    and build_version.
        """

    @abc.abstractmethod
    def flash_device(self, list_of_files,
                     expected_version=None,
                     expected_build_type=None,
                     verify_flash=True,
                     method=None):
        """Transfer a list of local files onto the device and flash the build on the device.

        Args:
            list_of_files (list): list of files on local computer.
            expected_version (str): firmware_version that is expected for the given list of files.
            expected_build_type (str): build_type that is expected for the given list of files.
            verify_flash (bool): True if flash_device should verify the firmware_version and
                                 build_type following the upgrade.
            method (str): upgrade method.

        Note:
            The list of files and expected version will be used to do an over the wire upgrade
            for the device.
        """

    @abc.abstractmethod
    def latest_verified_build_folder(self):
        """Returns the remote build folder path for the latest verified build.

        Returns:
            str: The latest verified remote build folder path.
        """

    @abc.abstractmethod
    def upgrade(self,
                build_number=None,
                build_url=None,
                build_file=None,
                forced_upgrade=False,
                latest_verified=False,
                **other_build_args):
        """Upgrade the device based on the provided build arguments.

        Args:
            build_number(int): build number. Defaults to last_believed good build.
            build_url(str): url or gsutil to the .zip file. Defaults to last_believed good build.
            build_file(str): local path to the file.
            forced_upgrade(bool): Upgrade the firmware without checks.
            latest_verified(bool): Upgrade to latest verified build.
            **other_build_args(dict): Other build arguments that are different based
                                      on the device type.
        """

    @abc.abstractmethod
    def upgrade_over_the_wire(self, **build_args):
        """Using the build arguments, download and flash the build on the device.

        Args:
            **build_args (dict): dictionary of build arguments.
        """



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for all raspbian devices."""
import re
import time

from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.base_classes import auxiliary_device
from gazoo_device.capabilities import file_transfer_scp
from gazoo_device.capabilities import shell_ssh
from gazoo_device.utility import host_utils

logger = gdm_logger.get_gdm_logger()

MARKER = "--- GDM Log Marker ---"
LOGGING_FILE_PATH = "/var/log/syslog"
LOG_MARKER_LINE_POS_OR_EMPTY = ('grep -n -e "{marker}" {file_path} --text | tail -n 1 | cut -d '
                                '":" -f 1').format(marker=MARKER, file_path=LOGGING_FILE_PATH)
LOG_MARKER_LINE_POS = ("line_num=$({cmd}); [ -z \"($line_num)\" ] && echo 1 || echo"
                       " $line_num".format(cmd=LOG_MARKER_LINE_POS_OR_EMPTY))

COMMANDS = {
    "BOOTUP_COMPLETE": "sudo systemctl --wait is-system-running",
    "FIRMWARE_VERSION": "cat /etc/os-release",
    "INJECT_LOG_MARKER": "sudo bash -c 'echo \"{marker}\" >> {file_path}'".format(
        marker=MARKER, file_path=LOGGING_FILE_PATH),
    "KERNEL_VERSION": "uname -r",
    "LOGGING": "tail -F -n +$({cmd}) {file_path}".format(cmd=LOG_MARKER_LINE_POS,
                                                         file_path=LOGGING_FILE_PATH),
    "MODEL_INFO": "cat /proc/device-tree/model",
    "GDM_HELLO": "echo 'GDM-HELLO'",
    "REBOOT": "sudo reboot",
    "RESET_FAILED": "sudo systemctl reset-failed",
    "SERIAL_NUMBER_INFO": "cat /proc/cpuinfo",
    "WEAVE_WBR_INSTALLED": "ls /usr/sbin/nldaemon-cli"
}

REGEXES = {
    "COMMAND_UNSUPPORTED": r"-bash: \S+: command not found",
    "device_type": "raspberrypi",
    "FIRMWARE_VERSION_REGEX": r"VERSION=\"(\d+ \(\w+\))\"",
    "KERNEL_VERSION_REGEX": r"(.*)",
    "MODEL_INFO_REGEX": r"Raspberry Pi ([^\n]+)",
    "SERIAL_NUMBER_INFO_REGEX": r"Serial\s+: ([^\n]+)"
}

TIMEOUTS = {
    "GDM_HELLO": 5,
    "SHELL": 10,
    "SHUTDOWN": 60,
    "ONLINE": 120
}


class RaspbianDevice(auxiliary_device.AuxiliaryDevice):
    """Base Class for Raspbian Devices."""

    COMMUNICATION_TYPE = "SshComms"
    _COMMUNICATION_KWARGS = {"log_cmd": COMMANDS["LOGGING"],
                             "ssh_key_type": "raspbian",
                             "username": "pi"}

    def __init__(self,
                 manager,
                 device_config,
                 log_file_name=None,
                 log_directory=None):
        super().__init__(manager,
                         device_config,
                         log_file_name=log_file_name,
                         log_directory=log_directory)
        self._commands.update(COMMANDS)
        self._regexes.update(REGEXES)
        self._timeouts.update(TIMEOUTS)

    @decorators.health_check
    def check_device_responsiveness(self):
        """Check if the device is responsive on console.

        Raises:
            DeviceNotResponsiveError: if device is not responsive on console.
        """
        cmd, timeout = self.commands["GDM_HELLO"], self.timeouts["GDM_HELLO"]
        try:
            self.shell(cmd, timeout=timeout)
        except errors.GazooDeviceError as err:
            raise errors.DeviceNotResponsiveError(
                self.name, "unable to execute command {!r} on device's shell".format(cmd),
                timeout=timeout, details=str(err))

    @decorators.DynamicProperty
    def kernel_version(self):
        """Version of Raspbian kernel.

        Returns:
            str: Raspbian kernel version.
        """
        return self.shell_with_regex(self.commands["KERNEL_VERSION"],
                                     self.regexes["KERNEL_VERSION_REGEX"],
                                     raise_error=True)

    @decorators.DynamicProperty
    def firmware_version(self):
        """Version of Raspbian.

        Returns:
            str: Raspbian version.
        """
        return self.shell_with_regex(self.commands["FIRMWARE_VERSION"],
                                     self.regexes["FIRMWARE_VERSION_REGEX"],
                                     raise_error=True)

    @decorators.PersistentProperty
    def health_checks(self):
        """Returns list of methods to execute as health checks."""
        return [self.device_is_connected,
                self.check_create_switchboard,
                self.check_device_responsiveness]

    @decorators.PersistentProperty
    def ip_address(self):
        """Global IP address."""
        return self.communication_address

    @decorators.CapabilityDecorator(shell_ssh.ShellSSH)
    def shell_capability(self):
        return self.lazy_init(shell_ssh.ShellSSH, self.switchboard.send_and_expect, self.name,
                              timeout=self.timeouts["SHELL"], tries=2)

    @decorators.CapabilityDecorator(file_transfer_scp.FileTransferScp)
    def file_transfer(self):
        """File transfer capability for moving files from and to the device.

        Returns:
            FileTransferScp: file transfer capability using "scp" command.
        """
        return self.lazy_init(file_transfer_scp.FileTransferScp,
                              ip_address_or_fn=self.ip_address,
                              device_name=self.name,
                              add_log_note_fn=self.switchboard.add_log_note,
                              user=self._COMMUNICATION_KWARGS["username"],
                              ssh_key_type=self._COMMUNICATION_KWARGS["ssh_key_type"])

    @decorators.LogDecorator(logger)
    def recover(self, error):
        """Recovers the device from an error detected by check_device_ready()."""
        if isinstance(error, errors.DeviceNotResponsiveError):
            self.reboot()
        else:
            raise error

    @decorators.LogDecorator(logger)
    def reboot(self, no_wait=False, method="shell"):
        """Reboots the device.

        Verifies device fully boots up afterwards.

        Args:
            no_wait (bool): Return before reboot completes. Default: False
            method (str): reboot technique to use.
        """
        self._inject_log_marker()
        self.switchboard.add_log_note("GDM triggered reboot")
        self.switchboard.send(command=self.commands["REBOOT"])
        if not no_wait:
            self._verify_reboot()

    @decorators.LogDecorator(logger)
    def get_detection_info(self):
        """Gets the persistent and optional attributes of a device during setup.

        Returns:
          tuple: (dict, dict) dictionary of persistent attributes,
                  dictionary of optional attributes (set to None).
        """
        self.props = self.props.copy()
        for cmd_name, command in self.commands.items():
            if not cmd_name.endswith("INFO"):
                continue
            regex = self._regexes[cmd_name + "_REGEX"]
            cmd_name = cmd_name.lower()[:-5]  # remove _INFO
            value = self.shell_with_regex(command, regex)
            if re.search(self.regexes["COMMAND_UNSUPPORTED"], value):
                value = "Raspbian detection did not support '{}'".format(command)
            self.props["persistent_identifiers"][cmd_name] = value

        return self.props["persistent_identifiers"], self.props["optional"]

    @classmethod
    def is_connected(cls, device_config):
        """Checks whether or not the device is connected to the computer.

        Args:
            device_config (dict): contains "persistent" dict

        Returns:
            bool: True if device is pingable, False otherwise.

        Notes:
            device_config is typically the device_config from manager.
            It checks if the device is reachable or not.
        """
        ip_address = device_config["persistent"]["console_port_name"]
        return host_utils.is_pingable(ip_address)

    def shell(self,
              command,
              command_name="shell",
              timeout=None,
              port=0,
              include_return_code=False):
        """Sends command and returns response and optionally return code.

        Args:
            command(str): Command to send to the device.
            command_name(str): Identifier for command.
            timeout(float): Time in seconds to wait for device to respond.
            port(int): Which port to send on, 0 or 1.
            include_return_code(bool): flag indicating return code should be returned.

        Raises:
            GazooDeviceError: if communication fails.

        Note:
            Can try multiple times as connection can sometimes fail.
            See shell_capability init args for setting the number of retry attempts.

        Returns:
            str: If include_return_code is False return the device response to the command.
            tuple: If include_return_code is True return the device response and return code.
        """
        timeout = timeout or self.timeouts["SHELL"]
        return self.shell_capability.shell(
            command, command_name=command_name, timeout=timeout, port=port,
            include_return_code=include_return_code)

    def shell_with_regex(self,
                         command,
                         regex,
                         regex_group=1,
                         command_name="shell",
                         raise_error=False,
                         tries=1,
                         port=0):
        """Sends a command, searches for a regex in the response, and returns a match group.

        Args:
            command(str): command to issue.
            regex(str): regular expression with one or more capturing groups.
            regex_group(int): number of regex group to return.
            command_name(str): command name to appear in log messages.
            raise_error(bool): whether or not to raise error if unable to find a match.
            tries(int): how many times to try executing the command before failing.
            port(int): which port to send the shell command to.

        Returns:
            str: value of the capturing group with index 'regex_group' in the match.

        Raises:
            GazooDeviceError: if command execution fails OR
                             couldn't find the requested group in any of the responses.
        """
        return self.command_with_regex(command,
                                       regex,
                                       self.shell,
                                       regex_group=regex_group,
                                       raise_error=raise_error,
                                       tries=tries,
                                       command_name=command_name,
                                       port=port)

    def _ensure_device_goes_offline(self, timeout=None):
        """Ensure device is no longer pingable over ssh.

        Args:
            timeout(float): Time in seconds to wait for device to respond.

        Raises:
            GazooDeviceError: Deviced failed to go offline before the timeout
        """
        timeout = timeout or self.timeouts["SHUTDOWN"]
        start_time = time.time()
        max_disconnect_time = start_time + timeout
        count = 0
        while time.time() < max_disconnect_time:
            if not host_utils.is_pingable(self.ip_address):
                count += 1  # Ensure device is really offline not just a blip
            else:
                count = 0
            if count == 2:
                logger.info("{} offline in {}s.".format(
                    self.name,
                    int(time.time() - start_time)))
                # close ssh transport as the ssh connection is disconnected.
                self.switchboard.close_all_transports()
                time.sleep(5)  # to ensure offline
                return
            time.sleep(.5)

        raise errors.GazooDeviceError(
            "Failed to go offline within {}s.".format(timeout))

    def _inject_log_marker(self):
        """Add GDM log marker to / var / log / syslog. This is to prevent reading stale logs.

        Note:
            Device logs are read starting with the last log marker(if present).
        """
        self.shell(self.commands["INJECT_LOG_MARKER"])

    def _ensure_device_is_online(self, timeout=None):
        """Ensure device is online and configs are fully loaded.

        Args:
            timeout(float): Time in seconds to wait for device to respond.

        Raises:
            GazooDeviceError: Device failed to come online before the timeout.
        """
        timeout = timeout or self.timeouts["ONLINE"]
        start_time = time.time()
        max_disconnect_time = start_time + timeout
        while time.time() < max_disconnect_time:
            if host_utils.is_pingable(self.ip_address):
                # There's a delay between the device being responsive to ping
                # and being able to open SSH connections
                time.sleep(10)
                self.switchboard.open_all_transports()
                break
            time.sleep(.5)
        else:
            raise errors.GazooDeviceError(
                "{} failed to become pingable in {}s.".format(self.name, timeout))

        output = "Device still offline"
        while time.time() < max_disconnect_time:
            # Ensure the BOOTUP_COMPLETE command is only sent once
            boot_up_complete_timeout = max_disconnect_time - time.time()
            try:
                output, return_code = self.shell(
                    self.commands["BOOTUP_COMPLETE"],
                    timeout=boot_up_complete_timeout,
                    include_return_code=True)
                if return_code == 0:  # command executed
                    logger.info(
                        "{} online in {}s".format(self.name, int(time.time() - start_time)))
                    return
                self.shell(self.commands["RESET_FAILED"])
            except errors.GazooDeviceError:
                logger.debug("{} failed to respond to {!r}.".format(
                    self.name, self.commands["BOOTUP_COMPLETE"]), exc_info=True)
            time.sleep(.5)
        raise errors.GazooDeviceError(
            "Failed to come online and respond to {!r} in {}s. Response: {}".format(
                self.commands["BOOTUP_COMPLETE"], timeout, output))

    def _verify_reboot(self):
        """Verifies reboot actually occurred."""
        self._ensure_device_goes_offline()
        self._ensure_device_is_online()

    def _list_properties_dynamic_raspbian(self):
        dyn_list = ["firmware_version",
                    "kernel_version"]
        return set(dyn_list)

    def __del__(self):
        self._log_object_lifecycle_event("__del__")
        if hasattr(self, 'close'):
            self.close()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Communication Power Ethernet Switch Capability."""
from gazoo_device import decorators
from gazoo_device import gdm_logger

from gazoo_device.capabilities.interfaces import comm_power_base

logger = gdm_logger.get_gdm_logger()


class CommPowerEthernetSwitch(comm_power_base.CommPowerBase):
    """Base class for comm_power_ethernet_switch capability."""

    def __init__(self,
                 device_name,
                 ethernet_switch_disable_fn,
                 ethernet_switch_enable_fn):
        """Create an instance of the comm_power_ethernet_switch capability.

        Args:
            device_name (str): name of the device this capability is attached to.
            ethernet_switch_disable_fn (func): disables ethernet_switch port for device's
                                               ethernet connection.
            ethernet_switch_enable_fn (func): enables ethernet_switch port for device's
                                              ethernet connection.
        """
        super().__init__(device_name=device_name)
        self._ethernet_switch_disable_fn = ethernet_switch_disable_fn
        self._ethernet_switch_enable_fn = ethernet_switch_enable_fn

    @decorators.CapabilityLogDecorator(logger)
    def off(self):
        """Turn off power to the device communications port."""
        self._ethernet_switch_disable_fn()

    @decorators.CapabilityLogDecorator(logger)
    def on(self):
        """Turn on power to the device communication port."""
        self._ethernet_switch_enable_fn()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Transport based on the websocket protocol. Implements GDM transport interface (TransportBase).

NOTE: the WebSocket protocol is NOT the same thing as a TCP/IP socket.
The WebSocket protocol is an application layer (layer 7) protocol.
    See https://tools.ietf.org/html/rfc6455.
TCP socket is a transport layer (layer 4) protocol.
    For a transport based on a TCP socket, see tcp_transport.py.
"""
import websocket
import traceback
from gazoo_device.switchboard import transport_base
from gazoo_device.switchboard import transport_properties as props
from gazoo_device.switchboard import tcp_transport
from gazoo_device import gdm_logger

logger = gdm_logger.get_gdm_logger(__file__)


class WebSocketTransport(transport_base.TransportBase):
    """WebSocket-based transport."""

    def __init__(self,
                 comms_address,
                 connect_timeout=tcp_transport.CONNECT_TIMEOUT,
                 auto_reopen=False,
                 open_on_start=True):
        """Initialize the WebSocketTransport.

        Args:
            comms_address (str): websocket URL to connect to (ws://...)
            connect_timeout (float): timeout on the socket instance before attempting connect.
            auto_reopen (bool): flag indicating transport should be reopened if
                                unexpectedly closed.
            open_on_start (bool): flag indicating transport should be open on
                                  TransportProcess start.
        """
        super(WebSocketTransport, self).__init__(auto_reopen, open_on_start)
        self._properties.update({
            props.CONNECT_TIMEOUT: connect_timeout,
            props.WEBSOCKET_URL: comms_address
        })
        self.comms_address = comms_address
        self._websocket = None

    def is_open(self):
        """Returns True if websocket has been created and is not closed.

        Returns:
            bool: True if _websocket is not None
        """
        return hasattr(self, "_websocket") and self._websocket is not None

    def _open(self):
        """Opens or reopens the websocket connection using the current property values.

        Raises:
            Exception: if connection could not be established.
        """
        self._websocket = websocket.create_connection(
            self._properties[props.WEBSOCKET_URL],
            self._properties[props.CONNECT_TIMEOUT])

    def _close(self):
        """Closes the websocket connection."""
        self._websocket.close()
        self._websocket = None

    def _read(self, size=None, timeout=None):
        """Receives data from the websocket.

        Args:
            size (int): not used.
            timeout (float): maximum seconds to wait to read bytes or indefinitely
                             if timeout=None. Default: None

        Returns:
            str: bytes read from transport
                 or "" if no bytes were read
                 or "" if an Exception occurred (including timeout)
        """
        self._websocket.settimeout(timeout)
        try:
            read_bytes = self._websocket.recv()
            return read_bytes
        except websocket.WebSocketTimeoutException:
            # Note: timeouts should not be logged. Otherwise, transport processes will flood
            # the logs with timeout exceptions (because they keep trying to read in infinite loop)
            return ""
        except Exception:
            logger.debug("_read() from websocket with URL {} failed due to an error."
                         .format(self._properties[props.WEBSOCKET_URL]))
            logger.debug(traceback.format_exc())
            return ""

    def _write(self, data, timeout=None):
        """Writes the data to the websocket.

        Args:
            data (str): bytes to be written out within timeout seconds
            timeout (float): maximum seconds to wait to write bytes or indefinitely
                             if timeout=None. Default: None

        Returns:
            int: number of bytes written (0 if any Exception occurs, including timeout)
        """
        self._websocket.settimeout(timeout)
        try:
            bytes_sent = self._websocket.send(data)
            return bytes_sent
        except websocket.WebSocketTimeoutException:
            # Note: timeouts should not be logged. Otherwise, transport processes will flood
            # the logs with timeout exceptions (because they keep trying to read in infinite loop)
            return 0
        except Exception:
            logger.debug("_write() to websocket with URL {} failed due to an error."
                         .format(self._properties[props.WEBSOCKET_URL]))
            logger.debug(traceback.format_exc())
            return 0



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common shell() capability for devices communicating over SSH."""
import time
from gazoo_device import config
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import shell_base

_SSH_CONNECTION_FAILURE_MARKERS = ["Connection to", "Connection reset"]

logger = gdm_logger.get_gdm_logger()


class ShellSSH(shell_base.ShellBase):
    """Common shell() method implementation for devices communicating over SSH."""

    def __init__(self, send_and_expect, device_name, shell_cmd=shell_base._SHELL_CMD,
                 shell_regex=shell_base._SHELL_REGEX, tries=shell_base._TRIES,
                 timeout=shell_base._TIMEOUT, failure_markers=None):
        """Initalize the SSH shell capability.

        Args:
            send_and_expect (method): bound send_and_expect method of the device class instance.
            device_name (str): name of the device using this capability.
            shell_cmd (str): return code wrapper around the shell command to execute.
            shell_regex (str): shell regex to use. Must contain two capturing groups:
                               one for the output and one for the return code.
            tries (int): how many times to try sending the shell command.
            timeout (float): shell timeout in seconds.
            failure_markers (list): list of markers (strings) indicating SSH connection failure.
        """
        super(ShellSSH, self).__init__(send_and_expect=send_and_expect, device_name=device_name,
                                       shell_cmd=shell_cmd, shell_regex=shell_regex, tries=tries,
                                       timeout=timeout)

        if failure_markers is None:
            failure_markers = _SSH_CONNECTION_FAILURE_MARKERS

        self._failure_markers = failure_markers

    def shell(self, command, command_name="shell", timeout=None, port=0,
              include_return_code=False, searchwindowsize=config.SEARCHWINDOWSIZE):
        """Sends command and returns response and optionally return code.

        Args:
            command (str): Command to send to the device.
            command_name (str): Identifier for command.
            timeout (float): Time in seconds to wait for device to respond.
            port (int): Which port to send on, 0 or 1.
            include_return_code (bool): flag indicating return code should be returned.
            searchwindowsize (int): Number of the last bytes to look at.

        Raises:
            GazooDeviceError: if communication fails.

        Note:
            Can try multiple times as connection can sometimes fail.
            See the init args for setting the number of retry attempts.

        Returns:
            str: If include_return_code is False return the device response to the command.
            tuple: If include_return_code is True return the device response and return code.
        """
        if timeout is None:
            timeout = self._timeout

        command = self._shell_cmd.format(cmd=command.rstrip())

        logger.debug("{} sending {!r} to generate {} in {}s on port {}",
                     self._device_name, command, command_name, timeout, port)

        for attempt in range(self._tries):
            response = self._send_and_expect(command, [self._shell_regex], timeout=timeout,
                                             port=port, searchwindowsize=searchwindowsize)
            if not response.timedout:
                break

            if (any(failure_marker in response.before for failure_marker in self._failure_markers)
                    and attempt < self._tries - 1):
                logger.warning("{}: SSH connection died with output {}. Trying again."
                               .format(self._device_name, response.before))
                # SSH connection died. Retry.
                time.sleep(.1)
            else:
                raise errors.GazooDeviceError("Device {} shell failed for command {!r}. "
                                              "Timed out waiting {}s for response. "
                                              "Shell output: {!r}.".format(
                                                  self._device_name, command, timeout,
                                                  response.before))

        result = response.match.group(1).strip()
        return_code = int(response.match.group(2))

        if include_return_code:
            return result, return_code
        else:
            return result

    def has_command(self, binary_name):
        """Returns if binary_name is installed on the device.

        Args:
            binary_name (str): name of the executable.

        Returns:
            bool: True if the executable is found on the device, False otherwise.
        """
        _, result_code = self.shell(f'which {binary_name}\n', include_return_code=True)
        return result_code == 0



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines a serial port transport by wrapping the Pyserial interface.

TcpTransport implements the standard GDM transport interface (TransportBase).
"""
from __future__ import absolute_import
import fcntl
import os
import serial
import time

from gazoo_device import gdm_logger
from gazoo_device.switchboard import transport_properties as props
from gazoo_device.switchboard import transport_base
import six


logger = gdm_logger.get_gdm_logger()
DEFAULT_BAUDRATE = 115200
REOPEN_TIMEOUT = 20


class SerialTransport(transport_base.TransportBase):
    """Communication transport for serial port."""

    def __init__(self,
                 comms_address,
                 baudrate=DEFAULT_BAUDRATE,
                 bytesize=serial.EIGHTBITS,
                 parity=serial.PARITY_NONE,
                 stopbits=serial.STOPBITS_ONE,
                 use_high_baudrate_flow_control=False,
                 auto_reopen=False,
                 open_on_start=True):
        """Initialize the SerialTransport object with the given serial properties.

        Args:
            comms_address (str): the path to the serial device to use
            baudrate (int): the initial baudrate serial property to use
            bytesize (int): the initial byte size serial property to use
            parity (str): the initial parity serial property to use
            stopbits (int): the initial stop bits serial property to use
            use_high_baudrate_flow_control (bool): whether to send control characters (XOFF, XON)
                                                   before & after writing to prevent
                                                   buffer issues at high baud rate (> 115200).
            auto_reopen (bool): flag indicating transport should be reopened if
                                unexpectedly closed.
            open_on_start (bool): flag indicating transport should be open on
                                  TransportProcess start.
        """
        super(SerialTransport, self).__init__(auto_reopen, open_on_start)
        self._max_read_errors = 3
        self._properties.update({
            props.BAUDRATE: baudrate,
            props.BYTESIZE: bytesize,
            props.PARITY: parity,
            props.STOPBITS: stopbits,
            props.XONXOFF: False,
            props.RTSCTS: False,
            props.DSRDTR: False,
            props.EXCLUSIVE: True,
            props.READ_REOPEN: True,
            props.USE_HIGH_BAUDRATE_FLOW_CONTROL: use_high_baudrate_flow_control
        })
        self.comms_address = comms_address
        self._read_errors = 0
        self._serial = serial.Serial()
        self._serial.port = comms_address

    def is_open(self):
        """Returns True if transport is currently open.

        Returns:
            bool: True if transport is currently open, False otherwise.
        """

        return hasattr(self, "_serial") and self._serial.isOpen()

    def _open(self):
        """Opens or reopens the serial port using the current property values.

        Raises:
            SerialException: if serial port path or properties are invalid.
        """

        self._serial.baudrate = self._properties[props.BAUDRATE]
        self._serial.bytesize = self._properties[props.BYTESIZE]
        self._serial.parity = self._properties[props.PARITY]
        self._serial.stopbits = self._properties[props.STOPBITS]
        self._serial.xonxoff = self._properties[props.XONXOFF]
        self._serial.rtscts = self._properties[props.RTSCTS]
        self._serial.dsrdtr = self._properties[props.DSRDTR]
        self._serial.exclusive = self._properties[props.EXCLUSIVE]
        self._serial.timeout = None
        self._serial.write_timeout = None
        self._serial.open()
        # Prevent inheritance of file descriptors to exec'd child processes [NEP-1852]
        fd = self._serial.fileno()
        flags = fcntl.fcntl(fd, fcntl.F_GETFD)
        fcntl.fcntl(fd, fcntl.F_SETFD, flags | fcntl.FD_CLOEXEC)

    def _close(self):
        """Closes the serial port."""
        # Prevent holding on to the exclusive lock [NEP-2473]
        fcntl.flock(self._serial.fileno(), fcntl.LOCK_UN)
        self._serial.close()

    def _read(self, size=1, timeout=None):
        """Returns bytes read up to max_bytes within timeout in seconds specified.

        Args:
            size (int): maximum number of bytes to read within timeout seconds
            timeout (float): maximum seconds to wait to read bytes or indefinitely
                             if timeout=None. Default: None

        Raises:
            SerialException: when read or open raises one

        Returns:
            str: bytes read from transport or None if no bytes were read
        """
        self._serial.timeout = timeout
        try:
            in_waiting = self._serial.in_waiting
            if in_waiting > 3072:
                logger.warning("Serial port input buffer size exceeds 3072 bytes. "
                               "Data will be lost if buffer size exceeds 4095 bytes. \n"
                               "Input buffer size: {}".format(in_waiting))
            result = self._serial.read(size=size)
            if self._read_errors:
                self._read_errors -= 1
        except serial.SerialException as err:
            self._read_errors += 1
            no_data = "read but returned no data" in repr(err)
            read_reopen = self._properties[props.READ_REOPEN]
            if self._read_errors > self._max_read_errors or not no_data or not read_reopen:
                raise
            logger.warn("Recovering from read returned no data, try {}/{}".
                        format(self._read_errors, self._max_read_errors))
            # Close and reopen the serial port to fix no data serial port errors
            self.close()
            end_time = time.time() + REOPEN_TIMEOUT
            while not os.path.exists(self._serial.port) and time.time() < end_time:
                time.sleep(0.001)
            self.open()
            # Recursively try again up to max_read_error attempts
            result = self._read(size=size, timeout=timeout)
        return result

    def _write(self, data, timeout=None):
        """Writes data provided within timeout in seconds specified.

        Args:
            data (str): bytes to be written out within timeout seconds
            timeout (float): maximum seconds to wait to write bytes or indefinitely
                             if timeout=None. Default: None

        Raises:
            SerialTimeoutException: in case timeout != None and timeout is exceeded
                                    when writing to device.

        Returns:
            int: number of bytes written or None if no bytes were written

        Note:
            If the data provided is unicode it will be encoded into UTF-8
            compatible bytes before being written to the serial port.
        """
        if isinstance(data, six.text_type):
            data = data.encode("utf-8", errors="replace")

        # Whether to use flow control to prevent buffer issues at high baudrate
        use_flow_control = (self._properties[props.BAUDRATE] > DEFAULT_BAUDRATE
                            and self._properties[props.USE_HIGH_BAUDRATE_FLOW_CONTROL])

        if use_flow_control:
            self.send_xoff()

        self._serial.write_timeout = timeout
        bytes_written = self._serial.write(data)

        if use_flow_control:
            self.send_xon()

        return bytes_written

    def set_property(self, key, value):
        """Sets the property specified to the value provided.

        Args:
            key (str): of the property to set
            value (object): of the property to set

        Raises:
            KeyError: when key property doesn't exist
        """
        super(SerialTransport, self).set_property(key, value)

        # Propagate property if device is already open
        if self._serial.isOpen():
            if key == props.BAUDRATE:
                self._serial.baudrate = value
            if key == props.BYTESIZE:
                self._serial.bytesize = value
            if key == props.PARITY:
                self._serial.parity = value
            if key == props.STOPBITS:
                self._serial.stopbits = value
            if key == props.XONXOFF:
                self._serial.xonxoff = value
            if key == props.RTSCTS:
                self._serial.rtscts = value
            if key == props.DSRDTR:
                self._serial.dsrdtr = value
            if key == props.EXCLUSIVE:
                self._serial.exclusive = value

    def flush_buffers(self):
        """Flushes all buffers on the serial device. Used during baudrate switching.

        Args: None

        Raises:
            SerialException: when pyserial fails to flush the buffers.

        Returns: None
        """
        self._serial.flush()
        self._serial.reset_input_buffer()
        self._serial.reset_output_buffer()

    def send_xon(self):
        """Sends flow control XON byte.

        Args: None

        Raises:
            SerialException: when pyserial fails to send the XON byte.

        Returns: None
        """
        self._serial.set_input_flow_control(enable=True)

    def send_xoff(self):
        """Sends flow control XOFF byte.

        Args: None

        Raises:
            SerialException: when pyserial fails to send the XOFF byte.

        Returns: None
        """
        self._serial.set_input_flow_control(enable=False)

    def send_break_byte(self):
        """Sends break byte (Ctrl-C). Not to be confused by the UART "break condition".

        Args: None

        Raises:
            SerialException: when pyserial fails to send the break byte.

        Returns: None
        """
        self._serial.write(six.b('\x03'))



"""Example of a primary device controller.

This example targets a generic Linux device accessible over SSH.
Assumptions for this device controller to work:
- The target device is running Linux.
- The device responds to ping and can be SSHed into.
- SSH does not require a password. You may need to set up passwordless
  SSH access via ssh-copy-id.
- The SSH username is "root". If not, change the value of SSH_USERNAME.

To try it out:
- Check out the gazoo-device repository:
  `git clone https://github.com/google/gazoo-device.git; cd gazoo-device`
- Connect a Linux device to your host and ensure that all of the
  assumptions above are satisfied.
- Comment out the 2 detection criteria in DETECT_MATCH_CRITERIA
  (see the instructions there) and save the file.
- Create a virtual environment:
  `python3 -m virtualenv gdm_venv`
- Install the gazoo-device from the checked out source code:
  `gdm_venv/bin/pip install -e ./`
- Try detecting your device:
  `gdm_venv/bin/python gdm detect --static_ips=<IP_of_device>`
- Try the GDM CLI with your device:
  ```
  gdm_venv/bin/python gdm man linuxexample
  gdm_venv/bin/python gdm devices
  # Use your device name (from output of "gdm devices") instead of "linuxexample-1234" below
  gdm_venv/bin/python gdm issue linuxexample-1234 --help
  gdm_venv/bin/python gdm issue linuxexample-1234 - shell "echo 'foo'"
  gdm_venv/bin/python gdm get-prop linuxexample-1234
  ```
"""
from gazoo_device import decorators
from gazoo_device import detect_criteria
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.base_classes import ssh_device

logger = gdm_logger.get_gdm_logger()

COMMANDS = {
    "FIRMWARE_VERSION": "uname -r",
    # Commands prefixed with "INFO_" will run during detection.
    # Their return values will be stored in the persistent config file.
    # The values will be available in
    # self.props["persistent_identifiers"][<property>].
    # In this case, the property name will be "hardware_architecture".
    "INFO_HARDWARE_ARCHITECTURE": "uname -m",
    # This only works on Raspberry Pi.
    "INFO_MODEL": "cat /sys/firmware/devicetree/base/model 2>/dev/null",
    # This only works on Raspberry Pi.
    "INFO_SERIAL_NUMBER": 'cat /proc/cpuinfo | grep "Serial" '
                          '| cut -d" " -f 2',
    "REBOOT": "reboot"
}
REGEXES = {}
TIMEOUTS = {}
SSH_USERNAME = "root"  # You may need to change this.


# Note: the device controller class name must not end in "Base" or
# "Device" to be automatically detected by GDM. Class names ending in
# "Base" or "Device" are interpreted as base classes.
class LinuxExample(ssh_device.SshDevice):
    """Example primary device controller for a generic Linux device."""
    _COMMUNICATION_KWARGS = {
        **ssh_device.SshDevice._COMMUNICATION_KWARGS,
        "username": SSH_USERNAME
    }
    # Device controllers must correspond to a single device type.
    # Therefore, detect criteria must uniquely identify a single device
    # type. I'm cheating in this example as it is generic.
    # To avoid interference with more specific Linux device controllers
    # (such as Raspberry Pi), this match is always set to False.
    # Comment out the 2 criteria below to enable detection of generic
    # Linux devices.
    DETECT_MATCH_CRITERIA = {
        # Comment these 2 lines out
        detect_criteria.SSH_QUERY.is_rpi: True,
        detect_criteria.SSH_QUERY.is_unifi: True
    }
    DEVICE_TYPE = "linuxexample"

    def __init__(self,
                 manager,
                 device_config,
                 log_file_name=None,
                 log_directory=None):
        super().__init__(manager,
                         device_config,
                         log_file_name=log_file_name,
                         log_directory=log_directory)
        self._commands.update(COMMANDS)
        self._regexes.update(REGEXES)
        self._timeouts.update(TIMEOUTS)

    @decorators.LogDecorator(logger)
    def factory_reset(self):
        """Resets the device to its factory default settings.

        Raises:
            NotImplementedError: there's no command to factory reset a
                                 generic Linux device.
        """
        raise NotImplementedError("Not really possible for a generic "
                                  "Linux device.")

    @decorators.DynamicProperty
    def firmware_version(self) -> str:
        """Returns the firmware version of the device."""
        return self.shell(self.commands["FIRMWARE_VERSION"])

    @decorators.LogDecorator(logger)
    def get_detection_info(self):
        """Gets the persistent and optional attributes of a device.

        Returns:
            tuple: (persistent properties dict, optional properties dict)
        """
        persistent_props, optional_props = super().get_detection_info()
        # Note that setting the "serial_number" persistent property is
        # required to generate the GDM name for the device.
        # If the device doesn't expose the serial number, set
        # self.props["persistent_identifiers"]["serial_number"] to any
        # unique identifier such as a MAC address of an interface.
        if not persistent_props.get("serial_number"):
            persistent_props["serial_number"] = "12345678"
        # Setting the "model" persistent property is also required.
        # Typical values are "Production", "Development", "Prototype".
        if not persistent_props.get("model"):
            persistent_props["model"] = "Production"

        return persistent_props, optional_props

    @decorators.PersistentProperty
    def hardware_architecture(self) -> str:
        """Returns the hardware architecture (such as "x86_64").

        This property is not required by GDM. It's an example of a
        persistent property which is populated during detection.
        """
        return self.props["persistent_identifiers"][
            "hardware_architecture"]

    @decorators.PersistentProperty
    def platform(self) -> str:
        """Returns the platform of the device."""
        return "new_top_secret_platform"

    @decorators.LogDecorator(logger)
    def reboot(self, no_wait=False, method="shell"):
        """Issues a soft reboot command.

        Args:
            no_wait (bool): flag indicating whether reboot verification
                            should be skipped. If False, blocks until
                            reboot completion.
            method (str): reboot technique to use.

        Raises:
            GazooDeviceError: if the provided reboot method is not supported.
        """
        if method not in ["shell"]:
            raise errors.GazooDeviceError(
                "{} reboot failed. Unsupported reboot method {!r} requested."
                .format(self.name, method))
        self.shell(self.commands["REBOOT"])
        if not no_wait:
            self._verify_reboot()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Dynamically generated maps of all classes supported by GDM.

These are populated by reflection_utils.py at import time (in gazoo_device/__init__.py).
These values are intended for internal GDM usage only.
See Manager.get_all_supported_<...> methods for the user-facing APIs returning the same data.
"""
mappings_generated = False
SUPPORTED_AUXILIARY_DEVICE_CLASSES = None
SUPPORTED_CAPABILITIES = None
SUPPORTED_CAPABILITY_INTERFACES = None
SUPPORTED_CAPABILITY_FLAVORS = None
SUPPORTED_PRIMARY_DEVICE_CLASSES = None
SUPPORTED_VIRTUAL_DEVICE_CLASSES = None



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""USB port map utility."""

from __future__ import absolute_import
import weakref

from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.auxiliary_devices import cambrionix
from gazoo_device.utility import usb_utils


logger = gdm_logger.get_gdm_logger()

# The PRODUCT_DEVICE_MAP is a dictionary with 'vendor_id/product_id' as the key
# and 'description' and 'type' in the values dict. The description includes
# the device_type in GDM as applicable.
# Example entry: {"1234/5678: {"description": "FooDevice", "type": "USB"}
PRODUCT_DEVICE_MAP = {}

# List of device types with a single serial connection.
SINGLE_CONNECTION_LIST = []


HUB_DEVICES_COLUMN_HEADER = "    Issue         Mode    Discovered                Configured"
OTHER_DEVICES_COLUMN_HEADER = "    Issue         Mode    Discovered"
OTHER_DEVICES_HEADER = "Other devices"


# Conflict types that port-map will test for and report.
USB_HUB_CONFLICT = "usb hub conflict"
USB_PORT_CONFLICT = "usb port conflict"
MISSING_USB_HUB = "missing usb hub"
MISSING_USB_PORT = "missing usb port"
NOT_DISCOVERED_CONFLICT = "not discovered"
NOT_DETECTED_CONFLICT = "not detected"


def _initialize_default_device(device):
    """Set values to the default for a device dictionary.

    Args:
        device (dict): The dictionary to be initialized with the default device data.
    """
    device['detected_devices'] = []
    device['device_type'] = None
    device['address'] = ''
    device['usb_hub_address'] = ''
    device['usb_hub_port'] = ''
    device['type'] = None
    device['serial_number'] = None
    device['conflict_types'] = []
    device['known_device'] = False
    device['description'] = '-'
    device['ftdi_interface'] = 0


def _identify_device(usb_info_inst):
    """Identify a udev usb device object and gather information about it.

    Args:
        usb_info_inst (UsbInfo): usb_info instance with usb information.

    Returns:
        dict: A dict containing device properties with the following keys:
            'description' (required): description of the Gazoo device, or contents of ID_SERIAL
                                if device not recognized.
            'name' (optional): GDM device name of the device, if recognized.
            'path_id' (required): the first part of the device id path.
            'part_id' (optional): the last part of the device id path.
    """
    product = "{}/{}".format(usb_info_inst.vendor_id, usb_info_inst.product_id)
    device_map_entry = PRODUCT_DEVICE_MAP.get(product)
    device_data = {}
    _initialize_default_device(device_data)
    if device_map_entry:
        device_data['known_device'] = True
        if 'device_type' in device_map_entry:
            id_suffix = usb_info_inst.serial_number[-4:].lower()
            device_data['name'] = "{}-{}".format(device_map_entry['device_type'], id_suffix)

        if 'type' in device_map_entry:
            device_data['type'] = device_map_entry['type']

        device_data['description'] = device_map_entry['description']
    else:
        device_data['description'] = usb_info_inst.serial_number

    device_data['serial_number'] = usb_info_inst.serial_number
    device_data['address'] = usb_info_inst.address
    device_data['usb_hub_address'] = usb_info_inst.usb_hub_address
    device_data['usb_hub_port'] = usb_info_inst.usb_hub_port
    device_data['ftdi_interface'] = usb_info_inst.ftdi_interface
    return device_data


def _get_usb_hub_map(device_info_list):
    """Creates a map of usb hub addresses to device_infos by port.

    Args:
        device_info_list (list): list of known usb_connections dicts.

    Returns:
       dict: map of usb hub addresses to device_infos by port
    """
    map_usb_hub_ports = {}
    for device_info in device_info_list:
        hub_address = device_info['usb_hub_address']
        port = device_info['usb_hub_port']
        if hub_address:
            if hub_address not in map_usb_hub_ports:
                map_usb_hub_ports[hub_address] = {}
            if not map_usb_hub_ports[hub_address].get(port) or device_info[
                    'ftdi_interface'] == 2:
                map_usb_hub_ports[hub_address][port] = device_info
    return map_usb_hub_ports


def _discover_cambrionix_devices(manager, cambrionix_info_list, device_info_list):
    """Discover all of the devices that are connected to a cambrionix.

    For each Cambrionix that is in the cambrionix_info_list, this function will use the devices
    in the device list to create port_list with information about each device that is connected
    to the Cambrionix USB hub.

    Args:
        manager (Manager): instance of Manager class
        cambrionix_info_list (list): list of devices for Cambrionix USB hubs
        device_info_list (list): list of devices for devices that have been detected

    Note:
        See the _identify_device function for more information on the device_data dicts that
        are used for the cambrionix_info_list and device_info_list.
    """
    device_config = {}
    map_usb_hub_ports = _get_usb_hub_map(device_info_list)
    for cambrionix_info in cambrionix_info_list:
        cambrionix_info['is_supported_hub'] = False
        try:
            device_config = {
                "persistent": {
                    "console_port_name": cambrionix_info['address'],
                    "name": "{}".format(cambrionix_info['name']),
                    "serial_number": cambrionix_info['serial_number']},
                "options": {},
                "make_device_ready": "on",
                "log_name_prefix": "port-map"
            }
            # Create the hub object. We don't use manager here since this is used
            # for hubs that have not been detected.
            cambrionix_inst = cambrionix.Cambrionix(manager=manager,
                                                    device_config=device_config,
                                                    log_file_name=None,
                                                    log_directory=manager.log_directory)
            cambrionix_inst.get_detection_info()  # populate undetected ones.
        except (KeyError, errors.GazooDeviceError):
            cambrionix_info['port_list'] = []
            cambrionix_info['port_count'] = 0
            continue

        try:
            cambrionix_info['port_count'] = cambrionix_inst.total_ports
            cambrionix_info['port_list'] = [None] * (cambrionix_info['port_count'] + 1)
            port_map = map_usb_hub_ports.get(cambrionix_info['address'], {})
            for port, device_info in port_map.items():
                cambrionix_info['port_list'][port] = device_info
            cambrionix_info['is_supported_hub'] = True
            cambrionix_info['port_modes'] = cambrionix_inst.switch_power.get_all_ports_mode()
        finally:
            cambrionix_inst.close()


class UsbPortMap(object):
    """Class for discovering USB information for connected devices, and printing a port-map."""

    def __init__(self, manager):
        """Initialize the USB port map instance by discovering devices and USB hubs.

        Args:
            manager (Manager):  instance of Manager class
        """
        self.manager_weakref = weakref.ref(manager)
        self.cambrionix_info_list = []
        self.device_info_list = []
        for usb_info_inst in usb_utils.get_address_to_usb_info_dict().values():
            device_info = _identify_device(usb_info_inst)
            if device_info:
                if 'cambrionix' in device_info.get('name', ''):
                    self.cambrionix_info_list.append(device_info)
                else:
                    self.device_info_list.append(device_info)

        _discover_cambrionix_devices(self.manager_weakref(),
                                     self.cambrionix_info_list,
                                     self.device_info_list)

        # remove devices associated with a cambrionix from the device list
        self.device_info_list = [device_info for device_info in self.device_info_list
                                 if not device_info['usb_hub_address']]

        # use what has been detected to add to what was discovered
        gdm_config = self.manager_weakref()._devices
        self._correlate_discovered_and_detected(gdm_config)
        self._get_detected_device_information(gdm_config)
        self._get_power_state_information()
        self._check_for_conflicts(gdm_config)

    def get_port_map(self):
        """Get a dictionary containing the port map information for the discovered devices.

        Returns:
            list: A list of dictionaries. Each entry in the list will be a hub or a device.
                  A hub will have a list of devices by port number, and a device will have the
                  port-map fields for that device, similar to what is printed by print_port_map.
        """
        port_map = []
        for cambrionix_info in self.cambrionix_info_list:
            port_map.append(self._get_hub_and_device_information(cambrionix_info))

        for device_info in self.device_info_list:
            other_info = self._get_other_device_information(device_info)
            if other_info:
                port_map.append(other_info)

        return port_map

    def print_port_map(self):
        """Print the port map based on what was discovered when the UsbPortMap object was created.
        """
        port_map = self.get_port_map()

        # Print the information for each hub in the port map
        for hub_or_device in port_map:
            if hub_or_device['hub_device']:
                hub_header_printed = False
                if 'port_list' not in hub_or_device:
                    continue
                for port, single in enumerate(hub_or_device['port_list']):
                    if port == 0 or not single:
                        continue
                    if not hub_header_printed:
                        hub_header_printed = True
                        if hub_or_device['is_detected']:
                            logger.info("\n{}".format(hub_or_device['hub_name']))
                        else:
                            logger.info("\n{} - Undetected".format(hub_or_device['hub_name']))
                        logger.info('-' * len(hub_or_device['hub_name']))
                        logger.info("\n{}\n".format(HUB_DEVICES_COLUMN_HEADER))

                    logger.info("{:>2}: {:13} {:7} {:25} {}".format(
                        port,
                        single['Issue'],
                        single['Mode'],
                        single['Discovered'],
                        single['Configured']))

        # Now that the hubs have been printed, loop through and print single devices
        other_header_printed = False
        for hub_or_device in port_map:
            if not hub_or_device['hub_device']:
                if not other_header_printed:
                    other_header_printed = True
                    logger.info("\n{}".format(OTHER_DEVICES_HEADER))
                    logger.info('-' * len(OTHER_DEVICES_HEADER))
                    logger.info("\n{}\n".format(OTHER_DEVICES_COLUMN_HEADER))

                logger.info("    {:13} Sync    {}".format(hub_or_device['Issue'],
                                                          hub_or_device['Discovered']))

    def _get_hub_and_device_information(self, hub):
        """Get information for all devices connected to a known USB hub.

        Args:
            hub (dict): A dictionary containing the information for the USB hub
        Returns:
            dict: A dictionary containing just the port map information for the hub, similar
                  to what is printed for print_port_map.
        """
        hub_dict = {}
        hub_dict['hub_name'] = hub['name']
        hub_dict['hub_device'] = True
        hub_dict['is_detected'] = hub['is_detected']
        if not hub['is_supported_hub']:
            hub_dict['port_count'] = 0
            hub_dict['port_list'] = []
            return hub_dict

        hub_dict['port_list'] = [None] * (hub['port_count'] + 1)
        hub_dict['port_count'] = hub['port_count']

        for port, device in enumerate(hub['port_list']):
            if port == 0:
                continue
            if device:
                # Based on the information on data in device, determine what will be printed
                mode_string = 'N/A'
                if 'port_mode' in device:
                    mode_string = device['port_mode']
                device_string = device['description']
                if 'gdm_config_name' in device:
                    device_string = device['gdm_config_name']
                if device['type']:
                    device_string += " " + device['type']
                configured_devices = ",".join(device['detected_devices'])
                issue_string = ""
                if configured_devices:
                    if len(device['detected_devices']) > 1:
                        issue_string = "Conflict"
                    if 'gdm_config_name' in device and \
                            device['gdm_config_name'] not in configured_devices:
                        issue_string = "Conflict"
                elif 'gdm_config_name' in device:
                    issue_string = "NotConfigured"
                if not issue_string and NOT_DETECTED_CONFLICT in device['conflict_types']:
                    issue_string = "Undetected"

                hub_dict['port_list'][port] = {}
                hub_dict['port_list'][port]['Issue'] = issue_string
                hub_dict['port_list'][port]['Mode'] = mode_string
                hub_dict['port_list'][port]['Discovered'] = device_string
                hub_dict['port_list'][port]['Configured'] = configured_devices
            else:
                logger.info("{:>2}: -".format(port))

        return hub_dict

    def _get_other_device_information(self, device):
        """Get the device information for a single device that is not associated with a USB hub.

        Args:
            device (dict): A dictionary containing information for a single device.
        Returns:
            dict: A dictionary containing device information for port-map or None.
        """
        if not device['known_device']:
            return None

        device_string = device['description']
        if 'gdm_config_name' in device:
            device_string = device['gdm_config_name']
        if device['type']:
            device_string += " " + device['type']
        issue_string = ""
        if USB_PORT_CONFLICT in device['conflict_types']:
            issue_string = "Conflict"
        if USB_HUB_CONFLICT in device['conflict_types']:
            issue_string = "Conflict"
        if NOT_DETECTED_CONFLICT in device['conflict_types']:
            issue_string = "Undetected"

        device_dict = {}
        device_dict['hub_device'] = False
        device_dict['Issue'] = issue_string
        device_dict['Mode'] = 'Sync'
        device_dict['Discovered'] = device_string
        device_dict['Configured'] = ''
        return device_dict

    def _check_known_devices(self,
                             device,
                             serial_numbers,
                             addresses):
        """Checks if usb connection matches a known device.

        Args:
            device (dict): A dictionary containing information about an individual device.
            serial_numbers (dict): The devices from GDM config with a configured serial number.
            addresses (dict): The devices from GDM config with a configured address.
        """
        if device:
            if device['address'] in addresses:
                device['gdm_config_name'] = addresses[device['address']]
            elif device['serial_number'] in serial_numbers:
                device['gdm_config_name'] = serial_numbers[device['serial_number']]
                del serial_numbers[device['serial_number']]

    def _correlate_discovered_and_detected(self, gdm_config):
        """Use the GDM config for detected devices to identify the discovered devices.

        Args:
            gdm_config (dict): a dictionary containing the devices detected by GDM.
        """
        serial_numbers = {}
        addresses = {}
        for name in gdm_config:
            if 'serial_number' in gdm_config[name]['persistent']:
                serial_numbers[gdm_config[name]['persistent']['serial_number']] = name
            if 'console_port_name' in gdm_config[name]['persistent']:
                addresses[gdm_config[name]['persistent']['console_port_name']] = name
        other_devices = self.manager_weakref().other_devices

        for cambrionix_info in self.cambrionix_info_list:
            cambrionix_info['is_detected'] = False
            if cambrionix_info['name'] in other_devices:
                cambrionix_info['is_detected'] = True

            for device_info in cambrionix_info['port_list']:
                self._check_known_devices(device_info,
                                          serial_numbers,
                                          addresses)

        for device_info in self.device_info_list:
            self._check_known_devices(device_info,
                                      serial_numbers,
                                      addresses)

        missing_device_info_list = []
        for number in serial_numbers:
            if serial_numbers[number] not in missing_device_info_list:
                missing_device_info_list.append(serial_numbers[number])

        for number in addresses:
            if addresses[number] not in missing_device_info_list:
                missing_device_info_list.append(addresses[number])

        # Some devices have a single connection, but may have an entry in more that one list.
        # For example, some devices will have a serial number and an adb serial number.
        # Iterate over the discovered devices and remove the extra entry so that they are not
        # reported as a missing device.
        for cambrionix_info in self.cambrionix_info_list:
            for device in cambrionix_info['port_list']:
                if not device:
                    continue

                if ('gdm_config_name' in device and
                        gdm_config[device['gdm_config_name']]['persistent']['device_type'] in
                        SINGLE_CONNECTION_LIST):
                    if device['gdm_config_name'] in missing_device_info_list:
                        missing_device_info_list.remove(device['gdm_config_name'])

        for device in self.device_info_list:
            if ('gdm_config_name' in device and
                    gdm_config[device['gdm_config_name']]['persistent']['device_type'] in
                    SINGLE_CONNECTION_LIST):
                missing_device_info_list.remove(device['gdm_config_name'])

        self.missing_devices = missing_device_info_list

    def _check_for_conflicts(self, gdm_config):
        """Use the GDM config for detected devices to add information or look for conflicts.

        Args:
            gdm_config (dict): a dictionary containing the devices detected by GDM
        """
        # Look for conflicts for devices that were discovered and are connected to a cambrionix
        for cambrionix_info in self.cambrionix_info_list:
            for index, device in enumerate(cambrionix_info['port_list']):
                if not device:
                    continue

                if device['known_device'] and 'gdm_config_name' not in device:
                    device['conflict_types'].append(NOT_DETECTED_CONFLICT)

                if 'gdm_config_name' in device:
                    device_id = device['gdm_config_name']
                    usb_hub_name, usb_port = self._get_usb_hub_info(device['type'],
                                                                    device_id,
                                                                    gdm_config)
                    if usb_hub_name and usb_hub_name != cambrionix_info['name']:
                        device['configured_hub'] = usb_hub_name
                        device['conflict_types'].append(USB_HUB_CONFLICT)

                    if usb_port and usb_port != str(index):
                        device['configured_port'] = usb_port
                        device['conflict_types'].append(USB_PORT_CONFLICT)

                    if device['type'] == 'USB':
                        if not usb_hub_name:
                            device['conflict_types'].append(MISSING_USB_HUB)
                        if not usb_port:
                            device['conflict_types'].append(MISSING_USB_PORT)

        # Look for conflicts for devices that were discovered and are not connected to a
        # cambrionix.
        for device in self.device_info_list:
            if device['known_device'] and 'gdm_config_name' not in device:
                device['conflict_types'].append(NOT_DETECTED_CONFLICT)
            if 'gdm_config_name' in device:
                device_id = device['gdm_config_name']
                usb_hub_name, usb_port = self._get_usb_hub_info(device['type'],
                                                                device_id,
                                                                gdm_config)
                if usb_hub_name:
                    device['conflict_types'].append(USB_HUB_CONFLICT)
                if usb_port:
                    device['conflict_types'].append(USB_PORT_CONFLICT)

    def _get_usb_hub_info(self,
                          connection_type,  # pylint: disable=unused-argument
                          device_id,
                          gdm_config):
        """Get the configured USB port for the specified device.

        Args:
            connection_type (str): no-op.
            device_id (str): the GDM name for a device.
            gdm_config (dict): a dictionary containing the devices detected by GDM.

        Returns:
            tuple: usb_hub_name, usb_port number. None if they are not set.
        """
        usb_hub_name = gdm_config[device_id]['persistent'].get('device_usb_hub_name')
        if not usb_hub_name:
            usb_hub_name = gdm_config[device_id]['options'].get('usb_hub')
        usb_port = gdm_config[device_id]['persistent'].get('device_usb_port')
        if not usb_port:
            usb_port = gdm_config[device_id]['options'].get('usb_port')
        return usb_hub_name, usb_port

    def _get_detected_device_information(self, gdm_config):
        """Use the GDM config for detected devices to add information to the device dictionary.

        Args:
            gdm_config (dict): a dictionary containing the devices detected by GDM
        """
        cambrionix_lookup = {}
        for index, cambrionix_info in enumerate(self.cambrionix_info_list):
            # usb_hub_name could be the cambrionix device name or the serial port path.
            cambrionix_lookup[cambrionix_info['address']] = index
            cambrionix_lookup[cambrionix_info['name']] = index

        # Add the detected device to the device at cambrionix/port if there is a mismatch
        for gdm_device in gdm_config:
            device_usb_hub_info = self._get_usb_hub_info('usb', gdm_device, gdm_config)
            for usb_hub_name, usb_port in [device_usb_hub_info]:
                if usb_hub_name and usb_port and usb_hub_name in cambrionix_lookup:
                    hub_index = cambrionix_lookup[usb_hub_name]
                    port_num = int(usb_port)
                    if port_num < len(self.cambrionix_info_list[hub_index]['port_list']):
                        if not self.cambrionix_info_list[hub_index]['port_list'][port_num]:
                            self.cambrionix_info_list[hub_index]['port_list'][port_num] = {}
                            device = self.cambrionix_info_list[hub_index]['port_list'][port_num]
                            _initialize_default_device(device)
                            device['conflict_types'].append(NOT_DISCOVERED_CONFLICT)
                            device['detected_devices'].append(gdm_device)
                        else:
                            device = self.cambrionix_info_list[hub_index]['port_list'][port_num]
                            device['detected_devices'].append(gdm_device)
                            if device.get('gdm_config_name') != gdm_device:
                                device['conflict_types'].append(NOT_DISCOVERED_CONFLICT)

    def _get_power_state_information(self):
        """For devices connected to a Cambrionix USB hub, get the power state mode.

        Note:
            This method should be called after _discover_cambrionix_devices. This method uses the
            the port state in the cambrionix device that is set by _discover_cambrionix_devices.
        """
        for cambrionix_info in self.cambrionix_info_list:
            if not cambrionix_info['is_supported_hub']:
                continue
            for index, device in enumerate(cambrionix_info['port_list']):
                if index == 0:
                    continue
                if not device:
                    cambrionix_info['port_list'][index] = {}
                    device = cambrionix_info['port_list'][index]
                    _initialize_default_device(device)

                device['port_mode'] = cambrionix_info['port_modes'][index]



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for GDM logger."""
from __future__ import absolute_import
import atexit
import logging
import logging.handlers
import os
import re
import sys
import types
import multiprocessing as mp
from gazoo_device import multiprocess_logging
from gazoo_device.config import DEFAULT_LOG_FILE
import six

# Making this global enables user control of stdout streaming (indirectly)
_stdout_handler = None

# Log formats for debug
FMT = ("%(asctime)s.%(msecs)03d %(levelname).1s %(process)5d "
       "%(filename)19.19s:%(lineno)d\t%(message)s")
DATEFMT = '%Y%m%d %X'

dirname = os.path.dirname(DEFAULT_LOG_FILE)
if not os.path.isdir(dirname):
    os.makedirs(dirname)


def add_handler(handler):
    """Adds a logging handler to the LoggingThread.

    It receives messages sent to GDM Loggers in the main process and its subprocesses.

    Along with remove_handler, allows library users to configure extra
    destinations and formats for log messages emitted from within GDM.

    Args:
        handler (logging.handler): A logging handler
    """
    logger = get_gdm_logger()
    logger.logging_thread.add_handler(handler)


def create_queue_handler(log_level):
    """Adds a QueueHandler to the top-level gazoo_device_manager logger.

    Should be used once per process, and will be the only handler on an GDM
    logger object in a given process. Destination handlers are added instead to
    the LoggingThread object via add_handler().

    Args:
        log_level (int): Integer log level, e.g. logging.DEBUG (value is 10)
    """
    logger = get_gdm_logger()
    handler = multiprocess_logging.QueueHandler(logger.logging_queue)
    handler.setLevel(log_level)
    logger.handlers = [handler]


def flush_queue_messages():
    """Wait until all messages currently in the logger queue are flushed."""
    get_gdm_logger().logging_thread.sync()


def get_gdm_logger(component_name=None):
    """Returns a Logger that inherits from (or is) the top-level GDM Logger.

    Differs from usual getLogger in that the name given is appended to the name
    of the top-level GDM Logger (e.g. get_gdm_logger('component') is equivalent
    to logging.getLogger('gazoo_device_manager.component')), and also in that
    message strings passed to the returned Logger will be combined with the
    given args using .format() rather than %.

    Args:
        component_name (str): name of an GDM component. Nests using '.' char.

    Returns:
      logging.Logger: main logger or sub logger.
    """
    if component_name is not None:
        name = '.'.join(['gazoo_device_manager', component_name])
    else:
        name = 'gazoo_device_manager'

    logger = logging.getLogger(name)

    # Override _log with our own _brace_format_log
    if not hasattr(logger, '_original_log'):
        logger._original_log = logger._log
        logger._log = types.MethodType(_brace_format_log, logger)

    return logger


def initialize_logger():
    """Configures the top-level GDM Logger.

    Configures it to begin logging to stdout and to the
    default log destination (e.g. /gazoo/gdm/log/gdm.txt).
    """
    # Set up multiprocess logging thread
    logging_queue = mp.Queue(-1)
    logging_queue.cancel_join_thread()
    logging_thread = multiprocess_logging.LoggingThread(logging_queue)
    atexit.register(logging_thread.stop)

    # Set up core GDM logger
    logger = get_gdm_logger()
    logger.setLevel(logging.DEBUG)
    logger.logging_queue = logging_queue
    logger.logging_thread = logging_thread

    # Create handler for core GDM logger in main process
    create_queue_handler(logging.DEBUG)

    # Configure a handler that writes to GDM logfile
    filepath = DEFAULT_LOG_FILE
    args = dict(mode='a', maxBytes=100 * 1024 * 1023, backupCount=5)
    logfile_handler = logging.handlers.RotatingFileHandler(filepath, **args)
    logfile_handler.setLevel(logging.DEBUG)
    logfile_formatter = logging.Formatter(FMT, datefmt=DATEFMT)
    logfile_handler.setFormatter(logfile_formatter)
    add_handler(logfile_handler)

    # Configure a handler that writes INFO logs to stdout
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setLevel(logging.INFO)
    stdout_formatter = logging.Formatter('%(message)s')
    stdout_handler.setFormatter(stdout_formatter)
    add_handler(stdout_handler)

    # Keep global copy of stdout_handler created and added above to allow for
    # changing the log level later for that handler.
    global _stdout_handler
    _stdout_handler = stdout_handler

    logging_thread.start()


def remove_handler(handler):
    """Removes the given handler from the GDM Logger's LoggingThread.

    Messages sent to the GDM Logger will no longer be sent to that handler.

    Along with add_handler, allows library users to configure extra destinations
    and formats for log messages emitted from within GDM.

    Args:
        handler (logging.handler): A logging handler that was added earlier using add_handler
    """
    logger = get_gdm_logger()
    logger.logging_thread.remove_handler(handler)


def set_component_log_level(component_name, log_level):
    """Sets the log level for the named component.

    Args:
        component_name (str): name of an GDM component
        log_level (int): Integer log level, e.g. logging.DEBUG (value is 10)
    """
    logger = get_gdm_logger(component_name)
    logger.setLevel(log_level)


def silence_progress_messages():
    """Stops the default behavior of streaming GDM Logger messages to stdout."""
    if _stdout_handler:
        remove_handler(_stdout_handler)


def stream_debug():
    """Sets the log level for GDM Logger stdout streaming to DEBUG."""
    if _stdout_handler:
        fmt = logging.Formatter(FMT, datefmt=DATEFMT)
        _stdout_handler.setFormatter(fmt)
        _stdout_handler.setLevel(logging.DEBUG)


def _brace_format_log(self, level, msg, args, exc_info=None, extra=None, **kwargs):
    """Enables brace logging for GDM Loggers (by overriding Logger._log)."""
    # Combines msg with args and kwargs using .format(),
    # saving special kwargs exc_info and extra for original _log method
    if args or kwargs:
        formatted_msg = str(msg).format(*args, **kwargs)
    else:
        formatted_msg = str(msg)

    new_kwargs = dict(exc_info=exc_info, extra=extra)

    self._original_log(level, formatted_msg, (), **new_kwargs)


class LogData(object):
    """Stores data to be added to log lines and automatically formatted."""

    key_pattern = re.compile(r'\W')

    def __init__(self, **items):
        self.items = items

    def __str__(self):
        r"""Formats data items into the form '##\tkey1=value1\tkey2=value2'.

        In keys, strips non-alphanumeric characters.
        In values, replaces escape char / and delimiters \t, = with /0, /1, /2.

        Returns:
          str: data as string.
        """
        strings = ['##']

        for k, v in six.iteritems(self.items):
            key = self.key_pattern.sub('', str(k))
            val = str(v).replace('/', '/0').replace('\t', '/1').replace('=', '/2')
            key_val_string = '='.join([key, val])
            strings.append(key_val_string)

        return '\t'.join(strings)
